// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// ProjectTo perform SA algorithms to train TDRF models
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


/**
 * \dir
 * \author wangbin
 * \date 2014-02-27
 * \todo 
 JRFmodelʹCDư취Ӧ
 * \brief JRFmodelʹCDư취Ӧ
*/
/**
 * \file
 * \author wangbin
 * \date 2014-02-27
 * \brief [main]
 */
#include "wbJRFCD.h"
#include "wbJRFSA.h"
#include "wbJRFClassModel.h"
#include "wbRosenfeld.h"
//#include <omp.h>
using namespace wbJRF;

static char *cfg_pathTrain = "LDC_gigaword_en.words.train.id";
static char *cfg_pathTest = "LDC_gigaword_en.words.test.id";
static char *cfg_pathValid = "LDC_gigaword_en.words.valid.id";
static char *cfg_pathVocab = NULL;
static char *cfg_pathTagVocab = NULL; //¼ÿtokenӦwordclass
static char *cfg_pathFeatures = NULL;
static char *cfg_pathInitModel = NULL; 
static char *cfg_pathModel = "JRFmodel_ML.txt";
static char *cfg_pathWriteEmpirical = NULL;
static char *cfg_pathReadEmpirical = NULL;
static char *cfg_pathReadFeatVar = NULL;
static char *cfg_pathWriteFeatVar = NULL;

static int cfg_nMaxLen = 25;
static char *cfg_strTitle = "SATrain";

static int cfg_nMinBatch = 3000;
static float cfg_fMinBatchStep = 0;
static int cfg_nSampleStep = 1;
static int cfg_nSamplePos = -1;
static int cfg_nIter_Beg = 0;
static float cfg_fIter_a0_lambda = 1.0;
static float cfg_fIter_a0_zeta = 1.0;
static float cfg_fIter_beta_lambda = 0;
static float cfg_fIter_beta_lambda2 = 1;
static float cfg_fIter_beta_zeta = 1;
static float cfg_fIter_beta_zeta2 = 1;
static float cfg_fIter_r_lambda = 1;
static float cfg_fIter_r_zeta = 1;
static float cfg_fIter_regular = 0;
static float cfg_fIter_regL1 = 0;
static float cfg_fIter_momentum = 0;
static bool cfg_bAverageLambda = false;
static bool cfg_bAverageZeta = false;
static int cfg_nAverageBeg = 1;
static bool cfg_bAutoTermate = false;
static float cfg_fAutoTerminalRate = 0.1;
static int cfg_dIter_t0 = 5000;
static int cfg_nIter_max = 10000;
static bool cfg_bIter_lbfgs = false;

static float cfg_fIter_hardThreshold = 0;
static float cfg_fIter_softThreshold = 0;

static int cfg_nPrintPerIter = 1; ///< ÿٴδӡһν

static bool cfg_bFastIter = false; ///< ٵвټһм
static int cfg_nThreadNum = 1;
//static bool cfg_bJumpNgram = false; ///< ¼ǷJumpNgram

static char *cfg_pathTempDir = NULL; ///< ʱļ·
static int cfg_nWriteTempModelPerIter = 0; ///< ÿٴεдһʱļ

static bool cfg_bUniformLenProb = 0; 

static char *cfg_pathDbgDir = NULL; ///< ļĿ¼

static int cfg_tconst = 0; 
//#ifdef Hessian
static float cfg_fHessianGap = 1e-5;
//#endif
#ifdef _LBFGS
static int cfg_nSetNum = 1; //ûٴεһs and y
static int cfg_nMemory = 10; 
#endif

#ifdef _Rosenfeld
static char *cfg_pathFakeset = NULL; 
#endif

static bool cfg_bRandomInit = false;
static float cfg_dAutoNormalize = 0;

wbOption options[] = {
	{wbOPT_STRING, "train", &cfg_pathTrain, "Train corpus (id-)"},
	{wbOPT_STRING, "test", &cfg_pathTest, "Test corpus (id-)"},
	{wbOPT_STRING, "valid", &cfg_pathValid, "Valid corpus (id-)"},
	{wbOPT_STRING, "vocab", &cfg_pathVocab, "ֵ,0ʼţʼ"},
	{wbOPT_STRING, "tag-vocab", &cfg_pathTagVocab, "ʹclassҪֵ"},
	{wbOPT_STRING, "feat", &cfg_pathFeatures, "ļ"},
	{wbOPT_STRING, "init-model", &cfg_pathInitModel, "ʼĵģ"},
	{wbOPT_STRING, "model", &cfg_pathModel, "ѵõģļ"},
	{wbOPT_INT, "max-len", &cfg_nMaxLen, "max-length of the JRF models"},
	{wbOPT_STRING, "title", &cfg_strTitle, "Show the title in Iteration"},
	{wbOPT_INT, "sample-step", &cfg_nSampleStep, "(default=1)"},
	{wbOPT_INT, "sample-pos", &cfg_nSamplePos, "Gibbs samplingٸλý"},
	{wbOPT_TRUE, "random-init", &cfg_bRandomInit, "Set random values for parameters"},

	{wbOPT_INT, "min-batch", &cfg_nMinBatch, "min-batch of Training(default=3000)"},
	{wbOPT_FLOAT, "min-batch-step", &cfg_fMinBatchStep, "increase the min-batch per iteration (default=0)"},
	{wbOPT_INT, "iter-beg", &cfg_nIter_Beg, "ӼʼĬΪ0"},

	{wbOPT_FLOAT, "iter-a0-lambda", &cfg_fIter_a0_lambda, "ʼѧϰ a0 for lambda"},
	{wbOPT_FLOAT, "iter-r-lambda", &cfg_fIter_r_lambda, "r for lambda"},
	{wbOPT_FLOAT, "iter-beta-lambda", &cfg_fIter_beta_lambda, "beta for lambda"},
	{wbOPT_FLOAT, "iter-beta-lambda2", &cfg_fIter_beta_lambda2, "(step2)beta for lambda"},
	{wbOPT_FLOAT, "iter-hard-threshold", &cfg_fIter_hardThreshold, "hard threshold (default = 0)"},
	{wbOPT_FLOAT, "iter-soft-threshold", &cfg_fIter_softThreshold, "soft threshold (default = 0)"},
	{wbOPT_FLOAT, "iter-a0-zeta", &cfg_fIter_a0_zeta, "ʼѧϰ a0 for zeta"},
	{wbOPT_FLOAT, "iter-r-zeta", &cfg_fIter_r_zeta, "r for zeta"},
	{wbOPT_FLOAT, "iter-beta-zeta", &cfg_fIter_beta_zeta, "beta for һϵzeta"},
	{wbOPT_FLOAT, "iter-beta-zeta2", &cfg_fIter_beta_zeta2, "(step2)beta for һϵzeta"},
	
	{wbOPT_FLOAT, "iter-regular", &cfg_fIter_regular, "ϵdefault=0"},
	{wbOPT_FLOAT, "iter-reg-l1", &cfg_fIter_regL1, "L1ϵdefault=0"},
	{wbOPT_FLOAT, "iter-momentum", &cfg_fIter_momentum, "momentumϵ, default=0"},
	{wbOPT_TRUE, "iter-avg-lambda", &cfg_bAverageLambda, "lambdaƽ"},
	{wbOPT_TRUE, "iter-avg-zeta", &cfg_bAverageZeta, "zetaƽ"},
	{wbOPT_INT, "iter-avg-beg", &cfg_nAverageBeg, "ƽʼ,Ӵ˴εʼȡƽ,ҪСt0"},
	{wbOPT_INT, "iter-t0", &cfg_dIter_t0, "˥ѧϰʣһt0"},
	{wbOPT_INT, "iter-max", &cfg_nIter_max, "(<0 ʾһֱ0ʾ>0ʾ)"},
	{wbOPT_TRUE, "iter-auto", &cfg_bAutoTermate, "Զt0ֹ"},
	{wbOPT_FLOAT, "iter-auto-rate", &cfg_fAutoTerminalRate, "ԶֹϵĬ0.1󲻳1"},
	{wbOPT_INT, "print-per-iter", &cfg_nPrintPerIter, "ÿٴδӡһνdefault=1"},
	{wbOPT_TRUE, "fast", &cfg_bFastIter, "ٵм̲һ"},
	{wbOPT_TRUE, "lbfgs", &cfg_bIter_lbfgs, "ʹţٷĬΪݶȷ"},
	{wbOPT_INT, "thread", &cfg_nThreadNum, "ʹ̲߳в߳"},
	{wbOPT_FLOAT, "auto-normalize", &cfg_dAutoNormalize, "ڹһ㣨0~10Լpipi඼Сֵֹ"},

	{wbOPT_INT, "tconst", &cfg_tconst, "˥ϵʱĶt"},
//#ifdef Hessian
	{wbOPT_FLOAT, "hgap", &cfg_fHessianGap, "СhgaphessianԪػᱻΪ0"},
//#endif
#ifdef _LBFGS
	{wbOPT_INT, "lbfgs-set", &cfg_nSetNum, "ÿٴθһsy"},
	{wbOPT_INT, "lbfgs-m", &cfg_nMemory, "LBFGSٸν"},
#endif

#ifdef _Rosenfeld
	{wbOPT_STRING, "fake-set", &cfg_pathFakeset, "Importance sampling Ҫһ"},
#endif

	{wbOPT_STRING, "temp-dir", &cfg_pathTempDir, "мʱģļ·"},
	{wbOPT_INT, "temp-iter", &cfg_nWriteTempModelPerIter, "ÿٴεдģļ(default=0,)"}, 

	{wbOPT_STRING, "write-empirical", &cfg_pathWriteEmpirical, "дÿľ"},
	{wbOPT_STRING, "read-empirical", &cfg_pathReadEmpirical, "ÿľ趨Զѵͳƣ"}, 
	{wbOPT_STRING, "read-var", &cfg_pathReadFeatVar, "Read Feature various File"},
	{wbOPT_STRING, "write-var", &cfg_pathWriteFeatVar, "Write Feature various File"},

	{wbOPT_TRUE, "uniform", &cfg_bUniformLenProb, "ʹþȳȷֲ"},
	{wbOPT_STRING, "dbg-dir", &cfg_pathDbgDir, "ļĿ¼"}
};

//#define TestGradient



_wbMain
{
	wbOpt_Parse(options);

	lout<<"*********************************************"<<endl;
	lout<<"          JRFSAmodel.exe { by Bin Wang }        "<<endl;
	lout<<"\t"<<__DATE__<<"\t"<<__TIME__<<"\t"<<endl;
	lout<<"**********************************************"<<endl;

#ifndef _DEBUG
	srand(time(0));
#endif

	//ompھļ
// 	omp_set_num_threads(cfg_nThreadNum);
// 	lout<<"[OMP] omp_thread = "<<omp_get_max_threads()<<endl;

	Vocab v(cfg_pathVocab);
	if (cfg_pathTagVocab)
		v.LoadVocabInfo(cfg_pathTagVocab);

#ifdef Scheme1
	#ifdef _ClassModel
		ClassModel_len m(&v, cfg_nMaxLen);
	#else
		CDmodel_len m(&v, cfg_nMaxLen);
	#endif  // end _ClassModel
		m.fileDbg.Open(wbString(cfg_pathModel).FileName()+".dbg", "wt");
#else // as Scheme 2
	#ifdef _ClassModel
		ClassModel_len m(&v, cfg_nMaxLen);
	#else
		CDmodel m(&v, cfg_nMaxLen);
	#endif // end _ClassModel
#endif // end Scheme1

	if (cfg_pathInitModel) {
		lout<<"Load Init model..."<<endl;
		m.ReadT(wbFile(cfg_pathInitModel, "rt"));
	} else if (cfg_pathFeatures) {
		lout<<"Load Features..."<<endl;
		m.ReadFeatures(wbFile(cfg_pathFeatures, "rt"));
	} else {
		lout_error("һЧʼģļ!");
	}
	m.PrepareInitModel(!cfg_bFastIter); //ģ׼ӹϣ
	m.m_opt_samplePos = cfg_nSamplePos; //gibbs samplingٸλý
	

	m.m_aLenProb.Clean();
	if (cfg_bUniformLenProb) {
		lout<<"Using Uniform Length Distribution"<<endl;
		m.m_aLenProb.SetNum(cfg_nMaxLen+1);
		m.m_aLenProb.Fill(1.0 / cfg_nMaxLen);
		m.m_aLenProb[0] = 0;
	}


#ifdef _Rosenfeld  //Rosenfeld
	ImSamplingFunc func(cfg_nMinBatch, cfg_fIter_regular, cfg_bFastIter);
	func.LoadFakeset(wbFile(cfg_pathFakeset, "rt"));
#else
	SA func(cfg_nMinBatch, cfg_fIter_regular, cfg_bFastIter);
#endif
	func.m_nSampleStep = cfg_nSampleStep;
	func.m_dL1Reg = cfg_fIter_regL1;
	func.m_pFileBeakDownLL = new wbFile(wbString(cfg_pathModel).FileName()+".ll", "wt");
	func.Prepare(cfg_nThreadNum, &m, cfg_pathTrain, cfg_pathValid, cfg_pathTest, 
		cfg_pathReadEmpirical, cfg_pathReadFeatVar, cfg_bUniformLenProb);
	//MCMC״̬
	func.m_pThreadMger->Load( wbFile(wbString(cfg_pathModel).FileName()+".mcmc", "rt", false) );
	if (cfg_pathWriteEmpirical) {
		lout<<"Write Empirical..."<<endl;
		func.WriteEmpirical(wbFile(cfg_pathWriteEmpirical, "wt"));
	}
	if (cfg_pathWriteFeatVar) {
		lout<<"Write Var..."<<endl;
		func.EmpiricalVar(func.m_corpusTrain);
		func.WriteEmpirical(wbFile(cfg_pathWriteFeatVar, "wt"));
	}

// 	int a[10] = {23, 17, 2};
// 	bool bFound;
// 	wbArray<int> **p = m.m_aGramIndex.Find(a, 3, bFound);

#ifdef TestGradient
	wbArray<double> aParam(func.m_nParamNum);
	wbArray<double> aExactGradient(func.m_nParamNum);
	wbArray<double> aApproGradient(func.m_nParamNum);
	for (int i=0; i<func.m_nParamNum; i++)
		aParam[i] = 1.2;

	double d0;
//  	d0 = func.GetValue(aParam);
//  	lout_variable(d0);
//  	lout_variable(func.m_pModel->m_logNormalizationFactor);


	func.m_pModel->SetParams(aParam, true);
	func.m_pModel->ExactNormalization();
	

	lout<<"Exacted..."<<endl;
	func.GetGradient(aParam, aExactGradient);

// 	lout<<"Approx..."<<endl;
// 	double dInc = 1e-3;
// 	titlePrecent(0, true, 3);
// 	for (int i=0; i<func.m_nParamNum; i++) {
// 		aParam[i] += dInc;
// 		double d1 = func.GetValue(aParam);
// 		/*lout_variable(d1-d0);*/
// 		aApproGradient[i] = (d1-d0)/dInc;
// 		aParam[i] -= dInc;
// 
// 		lout<<"["<<i<<"] "<<aExactGradient[i]<<"\t"<<aApproGradient[i]<<endl;
// 
// 		titlePrecent(i+1);
// 	}

	lout<<"Exacted..."<<endl;
	lout<<aExactGradient[0]<<endl;
	lout<<aExactGradient[1]<<endl;
	lout<<aExactGradient[2]<<endl;

// 	lout<<"Appro..."<<endl;
// 	lout<<aApproGradient[0]<<endl;
// 	lout<<aApproGradient[1]<<endl;
// 	lout<<aApproGradient[2]<<endl;

	wbFile f("gradient_SA.txt", "wt");
	for (int i=0; i<func.m_nParamNum; i++) {
		f.Print("%lf\n", aExactGradient[i]);
	}

	return 1;
#endif

	
	SAIter iter(&func, cfg_fIter_beta_lambda, cfg_fIter_beta_zeta, cfg_dIter_t0, 
		cfg_fIter_beta_lambda2, cfg_fIter_beta_zeta2,
		cfg_fIter_r_lambda, cfg_fIter_r_zeta );
	iter.m_dHardThreshold = cfg_fIter_hardThreshold;
	iter.m_dSoftThreshold = cfg_fIter_softThreshold;
	iter.m_nMinBatch_Init = cfg_nMinBatch;
	iter.m_nMinBatch_Step = cfg_fMinBatchStep;
	iter.m_opt_nIterBeg = cfg_nIter_Beg;
	iter.bLBFGS = cfg_bIter_lbfgs;
	//iter.fileParamDbg.Open(wbString(cfg_pathModel).FileName() + ".param", "wt");
	iter.a0_lambda = cfg_fIter_a0_lambda;
	iter.a0_zeta = cfg_fIter_a0_zeta;
	iter.fMomentum = cfg_fIter_momentum; ///< momentum
	iter.nAverageBeg = cfg_nAverageBeg;
	iter.bAverage_lambda = cfg_bAverageLambda;
	iter.bAverage_zeta = cfg_bAverageZeta;
	iter.bAutoTermate = cfg_bAutoTermate;
	iter.m_dTerminalRate = cfg_fAutoTerminalRate;
	iter.m_opt_nPrintPerIter = cfg_nPrintPerIter;
	iter.m_opt_nMaxIterNum = cfg_nIter_max;
	iter.m_strWriteModelDir = (cfg_pathTempDir)? cfg_pathTempDir: "";
	iter.m_nWriteModelPerIter = cfg_nWriteTempModelPerIter;
	iter.m_pPathModel = cfg_pathModel;
	iter.m_pfdbg_curHessian = (cfg_pathDbgDir)? new wbFile(wbString(cfg_pathDbgDir) + "\\" + wbString(cfg_pathModel).FileName() + ".curh", "wt"): NULL;
	if ( !cfg_bFastIter ) {
		iter.m_pfdbg_hessian = (cfg_pathDbgDir)? new wbFile(wbString(cfg_pathDbgDir) + "\\"+ wbString(cfg_pathModel).FileName() + ".hes", "wt"): NULL;
		iter.m_pfdbg_diffGradient = (cfg_pathDbgDir)? new wbFile(wbString(cfg_pathDbgDir) + "\\"+ wbString(cfg_pathModel).FileName() + ".diffG", "wt"): NULL;
		iter.m_pfdbg_dir = (cfg_pathDbgDir)? new wbFile( wbString(cfg_pathDbgDir) + "\\"+wbString(cfg_pathModel).FileName() + ".d", "wt" ) : NULL;
		iter.m_pfdbg_x = (cfg_pathDbgDir)? new wbFile( wbString(cfg_pathDbgDir) + "\\"+wbString(cfg_pathModel).FileName() + ".x", "wt" ) : NULL;
		iter.m_pfdbg_g = (cfg_pathDbgDir)? new wbFile( wbString(cfg_pathDbgDir) + "\\"+wbString(cfg_pathModel).FileName() + ".g", "wt" ) : NULL;
		iter.m_pfdbg_scaledG = (cfg_pathDbgDir)? new wbFile( wbString(cfg_pathDbgDir) + "\\"+wbString(cfg_pathModel).FileName() + ".sg", "wt" ) : NULL;
		iter.m_pfdbg_zeta = (cfg_pathDbgDir)? new wbFile( wbString(cfg_pathDbgDir) + "\\"+wbString(cfg_pathModel).FileName() + ".zeta", "wt" ) : NULL;
		iter.m_pfdbg_class = (cfg_pathDbgDir)? new wbFile( wbString(cfg_pathDbgDir) + "\\"+wbString(cfg_pathModel).FileName() + ".class", "wt" ) : NULL;
	}
	iter.m_tconstant = cfg_tconst;
	iter.m_dNormAutoRate = cfg_dAutoNormalize;
	//lout_variable(iter.m_dNormAutoRate);
	
	iter.m_dHessianGap = cfg_fHessianGap;
	iter.m_b0 = 0;
	lout_variable(iter.m_b0);

#ifdef _LBFGS
	iter.m_nLBFGS_avgNum = cfg_nSetNum;
	iter.m_nLimitiedNum = cfg_nMemory;
#endif

	lout<<"Iter Begin..."<<endl;
	if (cfg_strTitle)
		titlePrecent(0, 2, 0, cfg_strTitle);
	
	if (!cfg_pathInitModel && cfg_bRandomInit) {
		lout<<"[Random Init Values] for example:"<<endl;
		wbArray<double> aInitParams;
		srand(time(NULL));
		for(int i=0; i<func.m_nParamNum; i++)
			aInitParams[i] = 1.0 * rand() / RAND_MAX;

		for (int i=0; i<5; i++)
			lout<<"\t"<<aInitParams[i]<<endl;


		iter.Run(aInitParams);
	}	else {
		iter.Run(m.m_aValues);
	}

		

	lout<<"=== Iter End ==="<<endl;
	func.m_pModel->SetParams(iter.m_pdResParams, false);
	func.m_pModel->WriteT(wbFile(cfg_pathModel, "wt")); ///Ƚδһģдļ

	if ( cfg_bFastIter )
		return 1;

	lout<<"Caculate normalization...."<<endl;
	func.m_bFastIter = false;
	func.m_pModel->Reset(m.m_nOrder, m.m_nMaxLen);
	func.m_pModel->Normalization();
	func.m_pModel->WriteT(wbFile(cfg_pathModel, "wt"));
	
	double dExValue[10];

	func.GetExtraValues(NULL, dExValue);

	lout<<"logLikehood_train = "<<dExValue[0]<<endl;
	lout<<"logLikehood_test = "<<dExValue[1]<<endl;
	lout<<"PPL = "<<dExValue[2]<<endl;

	//Pause();


	return 1;
};