// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// Read TDRF models to calculate the PPL
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


/**
 * \dir
 * \author wangbin
 * \date 2014-03-29
 * \todo JRFģͣһЩ1һ2PPL or loglikelihood
 * \brief JRFģͣһЩ1һ2PPL or loglikelihood
*/
/**
 * \file
 * \author wangbin
 * \date 2014-03-29
 * \brief [main]
 */
#include "wbSystem.h"
#include "JRF\wbJRFCDmodel.h"
#include "JRF\wbJRFMLmodel.h"
using namespace wbJRF;

static char *cfg_pathModel = NULL;
static char *cfg_pathFeat = NULL;
static int cfg_nMaxLen = 0;
static char *cfg_pathWriteModel = NULL;
static char *cfg_pathVocab = NULL;
static char *cfg_pathTagVocab = NULL;
static char *cfg_strType = "scheme1";


static bool cfg_bNorm = false;
static bool cfg_bExact = false;
static int cfg_nSampleNum = 100000;
static int cfg_nHotNum = 1000;
static int cfg_nSampleStep = 2;
static float cfg_fThreshold = 0.1;

static char *cfg_pathTxt = NULL;
static char *cfg_pathOutResult = NULL;

wbOption options[] = {
	{wbOPT_STRING, "model", &cfg_pathModel, "[Input], model"},
	{wbOPT_STRING, "type", &cfg_strType, "[Input], ģ(scheme1 or scheme2),(default = scheme1)"},
	{wbOPT_STRING, "feat", &cfg_pathFeat, "[Input],ĲmodelļļҪֶָmaxLen"},
	{wbOPT_INT, "max-len", &cfg_nMaxLen, "[Input],ļʹ"},
	{wbOPT_STRING, "vocab", &cfg_pathVocab, "[Input], model vocabulary"},
	{wbOPT_STRING, "tag-vocab", &cfg_pathTagVocab, "[Input], model vocabulary containing class informations"},
	{wbOPT_STRING, "write-model", &cfg_pathWriteModel, "[Output], write model,һmodel"},
	// һҪĲ
	{wbOPT_TRUE, "norm", &cfg_bNorm, "[һ] ģͽйһ"},
	{wbOPT_TRUE, "exact", &cfg_bExact, "[һ] ȷһʹǰ㷨ͣʹýƹһ"},
	{wbOPT_INT, "sample", &cfg_nSampleNum, "[һ] "},
	{wbOPT_INT, "hot", &cfg_nHotNum, "[һ] Ԥȴ"},
	{wbOPT_INT, "step", &cfg_nSampleStep, "[һ] ȡֵ"},
	{wbOPT_FLOAT, "threshold", &cfg_fThreshold, "[һ] ֵֹ" },
	// PPL or likelihood
	{wbOPT_STRING, "txt", &cfg_pathTxt, "[PPL] PPL(id)"},
	{wbOPT_STRING, "out", &cfg_pathOutResult, "[PPL]  ӡÿϢ"}
};

_wbMain
{

	wbOpt_Parse(options);

	Vocab v(cfg_pathVocab);
	if ( cfg_pathTagVocab )
		v.LoadVocabInfo(cfg_pathTagVocab);

	CDmodel *pM = NULL;
	wbString strtype = cfg_strType;
	if (strtype == "scheme1")
		pM = new CDmodel_len(&v);
	else if (strtype == "scheme2")
		pM = new CDmodel(&v);
	//CDmodel m(&v);


	if (cfg_pathModel) {
		lout<<"Read model -> "<<cfg_pathModel<<endl;
		pM->ReadT(wbFile(cfg_pathModel, "rt"));
	} else if (cfg_pathFeat) {
		lout<<"Load feat -> "<<cfg_pathFeat<<endl;
		if (cfg_nMaxLen <= 0) {
			lout_error("Ϸmax-lenֵ!");
		}
		pM->ReadFeatures(wbFile(cfg_pathFeat, "rt"));
		pM->SetMaxLength(cfg_nMaxLen);
		cfg_bNorm = true; //< ڶļ˱ضйһ
	}

	//ķ
	int nonzero = 0;
	for (int i=0; i<pM->m_nFeatNum; i++) {
		if ( pM->m_aValues[i] != 0 ) 
			nonzero++;
	}
	lout<<"[Model Parameters] total="<<pM->m_nFeatNum<<" non-zero="<<nonzero<<"("<<100.0*nonzero/pM->m_nFeatNum<<"%)"<<endl;


	//
	if ( cfg_bNorm || cfg_pathTxt ) {
		pM->PrepareInitModel(cfg_bExact);
	}

	
	//һ
	if ( cfg_bNorm ) {

		if (cfg_bExact) {
			lout<<"Exact Normalization..."<<endl;
			pM->ExactNormalization();
		} else {
			lout<<"Approx Normalization..."<<endl;
			if (strtype == "scheme1") {
				wbArray<VocabID> aSeq(pM->m_nMaxLen+10);
				aSeq[0] = 0;
				int nLen = 1;
				titlePrecent(0, true, cfg_nHotNum, "[hot] ");
				for (int i=0; i<cfg_nHotNum; i++) {
					((CDmodel_len*)pM)->Sampling(aSeq, nLen, 1);
					titlePrecent(i+1);
				}
				titlePrecent(0, 2, 0, "[sample]");
				((CDmodel_len*)pM)->ApproxNormalization(aSeq, nLen, 0.1, cfg_nSampleNum, cfg_pathTxt);
			} else {
				pM->ApproxNormalization(cfg_nSampleNum, cfg_nHotNum, cfg_nSampleStep);
			}
		}

	}

	if ( cfg_pathTxt ) {
		double dPPL, dLogLikelihood;
		wbFile *pOut = NULL;
		if ( cfg_pathOutResult ) {
			pOut = new wbFile(cfg_pathOutResult, "wt");
		}
		pM->PPL(cfg_pathTxt, dPPL, dLogLikelihood, pOut);
		lout_variable(dPPL);
		lout_variable(dLogLikelihood);
		SAFE_DELETE(pOut);
	}

	if (cfg_pathWriteModel) {
		lout<<"Write model -> "<<cfg_pathWriteModel<<endl;
		pM->WriteT(wbFile(cfg_pathWriteModel, "wt"));
	}


	SAFE_DELETE(pM);

	return 1;
};