// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// TDRF model training method, for sampling
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.



/** 
 * \file
 * \author wangbin
 * \date 2014-03-17
 * \brief JRFģ, CDѵ㷨
 */

#pragma once
#include "wbJRFCDmodel.h"
#include "wbLBFGS.h"
#include "wbThread.h"

namespace wbJRF
{
	/**
	 * \class
	 * \brief ʹö߳ÿͳ
	 */
	class CDTaskPack : public wbTaskPack
	{
	public:
		// ڴʱ֪Ҫĸ̣߳ΪĿռ䣬ҪеĿռ䶼
		// ڵProcessʱݴ߳idȷʹĸݣʹĸ
		wbArray<VocabID> m_aSeq; ///< 
		int m_nSampleTimes; ///< 
		wbLHash<int,int> **m_paPeCount; ///< (ⲿ룩(¼ÿĳִϡԣʹùϣԼڴ
		wbLHash<int,int> **m_paPnCount; ///< (ⲿ룩Ĵ
		CDmodel *m_pMode;
		CDTaskPack(int nId, 
			wbLHash<int, int> **paPe, wbLHash<int,int> **paPn, CDmodel *pMode, 
			int nSampleTime)
			: wbTaskPack(nId), 
			m_paPeCount(paPe), m_paPnCount(paPn), 
			m_pMode(pMode), m_nSampleTimes(nSampleTime) {
				m_aSeq.SetNum(pMode->m_nMaxLen + 10);
		};
		virtual bool Process(int nThreadId);
	};
	/**
	 * \class
	 * \brief ̹
	 */
	class CDThreadMager : public wbThreadManager
	{
	public:
		wbLHash<int,int> **m_ppPeCount; ///< ÿ̵߳Ĺϣǰĸ
		wbLHash<int,int> **m_ppPnCount; ///< ÿ̵߳Ĺϣĸ

		int m_nSampleTime; ///< 
		CDmodel *m_pMode; ///< ģ
		Corpus *m_pTrain; ///< ѵָ
		int *m_pRandomSeq; ///< , ָָλãһȡm_nRandomNum
		int  m_nRandomNum; ///< ȡ
		int  m_nRandomCur; ///< ǰλ

	public:
		CDThreadMager(int nThreadNum, CDmodel *pMode, Corpus *pTrain, int *pRandSeq, int nRandNum, int nSampleTime)
			: wbThreadManager(nThreadNum), 
			m_pMode(pMode), m_pTrain(pTrain), 
			m_pRandomSeq(pRandSeq), m_nRandomNum(nRandNum), m_nRandomCur(0),
			m_nSampleTime(nSampleTime) 
		{
			//ٿռ
			m_ppPeCount = new wbLHash<int,int> *[nThreadNum];
			m_ppPnCount = new wbLHash<int,int> *[nThreadNum];
			for (int i=0; i<nThreadNum; i++) {
				m_ppPeCount[i] = new wbLHash<int,int>;
				m_ppPnCount[i] = new wbLHash<int,int>;
			}
		}
		~CDThreadMager()
		{
			//ͷſռ
			for (int i=0; i<m_nThreadNum; i++) {
				SAFE_DELETE(m_ppPeCount[i]);
				SAFE_DELETE(m_ppPnCount[i]);
			}
			SAFE_DELETE_ARRAY(m_ppPeCount);
			SAFE_DELETE_ARRAY(m_ppPnCount);
		}

		void Reset(int *pRandSeq, int nRandNum);
		virtual wbTaskPack* CreateTask();
		virtual void FinishTask( wbTaskPack *&pTask ) {
			titlePrecent(pTask->nId+1);
			delete (CDTaskPack*)pTask; //֣ò޷ͷڴĸо
			pTask = NULL;
		};
	};

	/** 
	 * \class
	 * \brief CD㷨
	 */
	class CD : public wbFunc
	{
	public:
		// 
		wbArray<int> m_aRandSeq; ///< Уᰴм¼˳ȡѵ
		int m_nBlockNum; ///< ѵϷֳɼblock
		int m_nBlockSize; ///< ÿblockжپ
		int m_nCurBlock; ///< ǰڼblock

		int m_nSamplingStep; ///< Pnеn

		double *m_pPefi; ///< ¼ÿľֵp~[f_i]
		double *m_pPnfi; ///< ÿμݶʱpn[fi]

		//
		Corpus m_corpusTrain;
		Corpus m_corpusTest;
		int m_nSeqMaxLen;		///< 󳤶
		int m_nSeqNum;			///< ѵϾ


		CDmodel *m_pModel; ///< ģ
		wbArray<int> m_aLenCount; ///< ¼ÿȵĸ棬ò

		bool m_bFastIter;  ///< ǷпٵCDһͺֵ
		int m_nScanTimes; ///< ϱ˶ٱ

		CDThreadMager *m_pThreadMager; ///< ̹߳ڼݶʱв

	public:
		CD(int nStepNum = 1, int nMinBatch = 3000, bool bFast = false):
		  m_nSamplingStep(nStepNum),m_nBlockSize(nMinBatch),m_nCurBlock(0),m_bFastIter(bFast),
			  m_pPefi(NULL), m_pPnfi(NULL),
			  m_nSeqMaxLen(0), m_nSeqNum(0), m_pModel(NULL)
		{
			m_nScanTimes = -1;
		}
		~CD()
		{
			m_pThreadMager->EndThread();
			SAFE_DELETE(m_pThreadMager);

			SAFE_DELETE_ARRAY(m_pPefi);
			SAFE_DELETE_ARRAY(m_pPnfi);
		}

		
		/// ѵ׼
		void Prepare(int nThreadNum, CDmodel *pModel, const char *pathTrain, const char *pathTest = NULL);

		/// 
		void RandTrainSeq();

		virtual bool GetGradient(double *pdParams, double *pdGradient)
		{
			return GetGradient_t(pdParams, pdGradient);
		}
		/// ߳ ݶ
		virtual bool GetGradient_t(double *pdParams, double *pdGradient);

		/// ̼߳ݶȣԱ
		virtual bool GetGradient_norm(double *pdParams, double *pdGradient)
		{
			//wbFile test("cd_gradient.dbg", "wt");

			if (pdParams) {
				m_pModel->SetParams(pdParams, false);
			}

			int nFeatrueNum = m_pModel->m_nFeatNum;
			memset(m_pPefi, 0, sizeof(m_pPefi[0])*nFeatrueNum);
			memset(m_pPnfi, 0, sizeof(m_pPnfi[0])*nFeatrueNum);

			if (m_nCurBlock >= m_nBlockNum)
				RandTrainSeq(); //

			int nBeg = m_nCurBlock * m_nBlockSize;
			int nEnd = min(nBeg + m_nBlockSize, m_nSeqNum);
			int nNum = nEnd-nBeg;
			
			wbArray<int> aSeq(m_pModel->m_nMaxLen + 10);

			wbFile test("testThread_s.dbg", "wt");
#ifdef wsj
			titlePrecent(0, true, nNum, "Gradient");
#endif // wsj
			for (int nRad=nBeg; nRad<nEnd; nRad++) { ///< ÿ
				
				//aSeqs = m_aTrainSeqs[m_aRandSeq[nRad]];
				m_corpusTrain.GetLine(m_aRandSeq[nRad], aSeq); ///õm_aRandSeq[nRad]е
				int nLen = aSeq.GetNum();

				wbArray<int> aSrc = aSeq;
				int nSrcLen = nLen;
#ifdef wsj
				lout<<"Sampling Begin ["<<nRad-nBeg<<"]"<<endl;
#endif
				for (int i=0; i<nSrcLen; i++)
					test.Print("%c", aSeq[i]+'a');
				test.Print("\t");

				m_pModel->Sampling(aSeq, nLen, m_nSamplingStep);
// 				srand(0);
// 				lout<<rand()<<endl;
				int nSamLen = nLen;

				for (int i=0; i<nSamLen; i++)
					test.Print("%c", aSeq[i]+'a');
				test.Print("\n");

				for (int i=0; i<nFeatrueNum; i++) {  ///< ÿ

					m_pPefi[i] += m_pModel->m_aFeatures[i]->nAccept( aSrc, nSrcLen );

					m_pPnfi[i] += m_pModel->m_aFeatures[i]->nAccept( aSeq, nSamLen );
				}

#ifdef wsj
				titlePrecent(nRad+1-nBeg);
#endif // wsj
			}

			for (int i=0; i<nFeatrueNum; i++) {
				m_pPefi[i] /= nNum;
				m_pPnfi[i] /= nNum;
			}

			m_nCurBlock++;

			//ݶ
			for (int i=0; i<m_nParamNum; i++) {
				pdGradient[i] = m_pPnfi[i] - m_pPefi[i];
			}

// 			for (int i=27; i<30; i++)
// 				lout<<m_pPnfi[i]<<"\t"<<m_pPefi[i]<<endl;

			return true;
		}
		
		virtual double GetValue(double *pdParams)
		{
#ifdef Scheme1
				return 0;
#endif
			if (pdParams)
				m_pModel->SetParams(pdParams);

			return CaculateLogLikelihood(&m_corpusTrain);
		}

		virtual int GetExtraValues(double *pdParams, double *pdValues)
		{
			
			static int t = -1;
			t++;


			if (pdParams)
				m_pModel->SetParams(pdParams, !m_bFastIter);

			//ѵȻֵ
			pdValues[0] = CaculateLogLikelihood(&m_corpusTrain, true);

			//ԼȻֵ, PPL
			pdValues[1] = CaculateLogLikelihood(&m_corpusTest, true, &pdValues[2]);

			//return 3;

#ifdef Scheme1
			//ʹþȷĹһϵȻֵ
			CDmodel_len *pModel = (CDmodel_len*)m_pModel;

			wbArray<double> aSave(m_pModel->m_nMaxLen); //zeta
			aSave.Copy(pModel->m_aApproxLogNormFactors);

			for (int i=1; i<=m_pModel->m_nMaxLen; i++) {
				pModel->m_aApproxLogNormFactors[i] = pModel->m_aLogNormFactors[i] - pModel->m_aLogNormFactors[pModel->j_zero];
			}

			//ԼȻֵ, PPL
			pdValues[3] = CaculateLogLikelihood(&m_corpusTest, true, &pdValues[4]);

			pModel->m_aApproxLogNormFactors.Copy(aSave);

			//ӡֲֵ
			pdValues[5] = ( pdValues[5] * t + pdValues[1] ) / (t+1);

			return 6;
#else
			return 3;
#endif
			
		}

		/// ѵ֤ϵȻֵ
		double CaculateLogLikelihood(Corpus *pCorpus, bool bNorm = true, double *pPPL=NULL);
	};


	/**
	 * \class
	 * \brief CD㷨ĵΪa0/(1+n/r), nΪrΪϵ
	 */
	class CDIter : public wbLBFGS
	{
	public:

		double a0_lambda;///< ʼѧϰ
		double a0_zeta; ///< ʼѧϰ
		double beta_lambda; ///< lambdabeta
		double beta_zeta;  ///< һϵbeta
		double beta_lambda2;  ///< step2beta
		double beta_zeta2;    ///< step2beta

		double r_lambda;
		double r_zeta;

		double gamma_lambda; ///< lambdaѧϰϵ
		double gamma_zeta; ///<  һϵѧϰϵ

		bool bAverage_lambda; ///< lambdaǷƽ
		bool bAverage_zeta;	  ///< zetaǷƽ
		int nAverageBeg; ///< ӦСt0;
		int nAverageEnd; ///< the end iteration for averaging.
		
		double *pAvgLambda; ///< ƽLambda

		double fMomentum; ///< momentumϵ
		double t0;
		//double dAverageValue;
		bool bAutoTermate; ///< ԶStepֹ

		wbArray<double> dAverageQueue; ///< ¼Ȼֵѭ
		int nAverageQueueLen; ///< г
		int nAverageQueueTop; ///< ж

		bool bLBFGS;
		//wbFile fileParamDbg; ///< ӡļ

		
		int m_nWriteModelPerIter;  ///< ٲ¼һεǰģ
		wbString m_strWriteModelDir; ///< ǰļдʱģ

		int m_nMonitorOverfittingPerIter; ///< ٲһѵͲԼϵȻֵ

	public:
		CDIter(wbFunc *pfunc, double p_beta_lambda, double p_beta_zeta, double p_t0, 
			double p_beta_lambda2 = 1, double p_beta_zeta2 = 1, double p_r_lambda=1, double p_r_zeta=1): wbLBFGS(pfunc)
		{
			a0_lambda = 1;
			a0_zeta = 1;
			beta_lambda = p_beta_lambda;
			beta_zeta = p_beta_zeta;
			t0 = p_t0;
			beta_lambda2 = p_beta_lambda2;
			beta_zeta2 = p_beta_zeta2;
			r_lambda = p_r_lambda;
			r_zeta = p_r_zeta;
			//m_opt_dFixedStep = 0.0;
			//dAverageValue = 0;
			fMomentum = 0;

			bAverage_lambda = false;
			bAverage_zeta = false;
			nAverageBeg = 1;
			nAverageEnd = INT_MAX;
			SAFE_NEW_ARRAY(pAvgLambda, double, pfunc->m_nParamNum);
			memset(pAvgLambda, 0, sizeof(double)*(pfunc->m_nParamNum));  //ֵΪ0

			nAverageQueueLen = 10;
			nAverageQueueTop = 0;
			dAverageQueue.SetNum(nAverageQueueLen);
			dAverageQueue.Fill(0);

			bAutoTermate = false;

			wbLBFGS::m_opt_bGradientDescent = false;

			m_nWriteModelPerIter = 0;
			m_nMonitorOverfittingPerIter = 10;
		}
		~CDIter()
		{
			SAFE_DELETE_ARRAY(pAvgLambda);
		}

		virtual void ComputeDir(int k, double *pdGradient, double *pdAlpha, double *pdDir)
		{
			int nVecLen = m_pFunc->m_nParamNum;
			double dScale = 1.0;

			if (bLBFGS && k+1 <= t0) { ///< ʹLBFGS㷽
				wbLBFGS::ComputeDir(k, pdGradient, pdAlpha, pdDir);
			} 




			if (k == 0) {
				for (int n=0; n<nVecLen; n++)
					pdDir[n] = dScale * ( -pdGradient[n] );  //Ҫȡ
			} else {
				for (int n=0; n<nVecLen; n++)
					pdDir[n] = dScale * ( -pdGradient[n] + fMomentum * pdDir[n] ); ///ʷ
			}
		}

		virtual double LineSearch(double *pdDir, double dValue, double *pdGradient, double *pdCurParam, double *pdNextParam)
		{
			int t = m_nIterNum+1; //1ʼ

			if (bAutoTermate) //ƽȻֹ
			{
				double dLastAverageValue = dAverageQueue[(nAverageQueueTop+nAverageQueueLen-1)%nAverageQueueLen]; //һ
				double dNewAverageValue = dLastAverageValue + 1.0/(t) * (dValue - dLastAverageValue); //Ŀǰµ
				double dOldAverageValue = dAverageQueue[(nAverageQueueTop+nAverageQueueLen-5)%nAverageQueueLen];
				if (t < t0) {
					if ( fabs(dNewAverageValue - dOldAverageValue) < 1e-5 ) { //ǰstep2
						t0 = t;
						lout<<"[SA] change t0="<<t0<<endl;
					}
				} 
				if ( t > t0 ) {
					if ( fabs(dNewAverageValue - dOldAverageValue) < 1e-5 ) { //ǰ
						m_opt_bEndIter = true;
						lout<<"[SA] Terminate (t0="<<t0<<",tmax="<<t<<")"<<endl;
					}
				}
				dAverageQueue[nAverageQueueTop] = dNewAverageValue; //
				nAverageQueueTop = (nAverageQueueTop+1)%nAverageQueueLen;
			}

			//ѧϰ
			if (t <= t0) {
				gamma_lambda = a0_lambda / pow( (1+t*r_lambda), beta_lambda );
				gamma_zeta = a0_zeta / pow( (1+t*r_zeta), beta_zeta );
			} else {
				//ڶҪͳzetalambdaĹ
				gamma_lambda = 0; // a0_lambda / ( pow( ((t-t0)*r_lambda), beta_lambda2) + pow( (1+t0*r_lambda), beta_lambda) );
				gamma_zeta = a0_zeta / ( pow( ((t-t0)*r_zeta), beta_zeta2) + pow( (1+t0*r_zeta), beta_zeta) );
			}

			lout<<"[SAIter] "<<"t="<<t<<" gamma_lambda="<<gamma_lambda<<" gamma_zeta="<<gamma_zeta<<endl;

			//²
			if ( bAverage_lambda && t <= t0 && t>= nAverageBeg ) { // ڶ׶βƽ
				if ( t == nAverageBeg ) { //ʼ
					memcpy(pAvgLambda, pdCurParam, sizeof(double)*(m_pFunc->m_nParamNum));
				}
				for (int i=0; i<m_pFunc->m_nParamNum; i++) {
					pAvgLambda[i] += gamma_lambda * pdDir[i];
				}
				for (int i=0; i<m_pFunc->m_nParamNum; i++) {
					pdNextParam[i] = pdCurParam[i] + 1.0/(t-nAverageBeg+1) * (pAvgLambda[i] - pdCurParam[i]);
				}

			} else {
				for (int i=0; i<m_pFunc->m_nParamNum; i++) {
					pdNextParam[i] = pdCurParam[i] + gamma_lambda * pdDir[i];
				}
			}

// 			if ((m_nIterNum+1)%m_opt_nPrintPerIter==0) {
// 				//ӡ
// 				for (int i=0; i<min(2000,m_pFunc->m_nParamNum); i++) {
// 					fileParamDbg.Print("%.2lf ", pdNextParam[i]);
// 				}
// 				fileParamDbg.Print("\n");
// 			}


#ifdef Scheme1
			CDmodel_len *pModel = (CDmodel_len *) ((CD*)m_pFunc)->m_pModel;

			pModel->UpdateApproximateValues(gamma_zeta, ((bAverage_zeta&&t<=t0&&t>=nAverageBeg)?t-nAverageBeg+1:-1));

			if ((m_nIterNum+1)%m_opt_nPrintPerIter==0) {
				pModel->OutDbg();
			}

#endif

			return gamma_lambda;	


// 			m_opt_dFixedStep = m_dStepInit / (1 + m_dr * m_nIterNum);
// 
// 			for (int i=0; i<m_pFunc->m_nParamNum; i++)
// 				pdNextParam[i] = pdCurParam[i] + m_opt_dFixedStep * pdDir[i];
// 
// 			return m_opt_dFixedStep;	
		}

		virtual void IterStepEnd(int k)
		{
			CD *pCdFunc = (CD*)m_pFunc;
			//ӡļ
			if ( m_nWriteModelPerIter>0 && (k+1)%m_nWriteModelPerIter == 0 ) 
			{
				char fileName[2048];
				if (m_strWriteModelDir != "" && *m_strWriteModelDir.End() != '\\')
					m_strWriteModelDir += "\\";
				sprintf(fileName, "%sTemp%d.mode", m_strWriteModelDir.GetBuffer(), (k+1)/m_nWriteModelPerIter);
				
				lout<<"[Write Model] => "<<fileName<<endl;
				pCdFunc->m_pModel->WriteT(wbFile(fileName, "wt"));
			}

			//ѵ֤ϵȻֵ
// 			if ( m_nMonitorOverfittingPerIter>0 && (k+1)%m_nMonitorOverfittingPerIter == 0 ) {
// 				double dTrain = pCdFunc->CaculateLogLikelihood( &(pCdFunc->m_corpusTrain), false );
// 				double dVaild = pCdFunc->CaculateLogLikelihood( &(pCdFunc->m_corpusTest), false );
// 
// 				lout<<"[Monitor] "<<"train="<<dTrain<<" vaild="<<dVaild<<" diff="<<dVaild-dTrain<<endl;
// 			}
			
		}
	};
}