// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// TDRF models, including sampling algorithms.
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


/** 
 * \file
 * \author wangbin
 * \date 2014-03-17
 * \brief JRFģ, CDѵ˰㷨
 */
#pragma once
#include "wbJRFmodel.h"
#include "wbJRFMLmodel.h"


namespace wbJRF
{
	

	/** 
	 * \class
	 * \brief CDƵJRFģͣʹýg(x|y)
	 */
	class CDmodel : public MLmodel
	{
	public:
		Prob **m_pLenJumpProb; ///< ģͳȵת

		int m_opt_samplePos; ///< ÿGibbs samplingλĿ, -1 means full sampling

	public:
		CDmodel(Vocab *pVocab, int nMaxLen = 10):MLmodel(pVocab, nMaxLen) {

			m_pLenJumpProb = NULL;
			SetMaxLength(nMaxLen);

			m_opt_samplePos = -1;
		}
		~CDmodel()
		{
			SAFE_DELETE_DARRAY(m_pLenJumpProb, m_nMaxLen+1);
		}
		/// ٳתռ
		virtual void SetMaxLength(int nMaxLen)
		{
			m_nMaxLen = nMaxLen;
			SAFE_NEW_DARRAY(m_pLenJumpProb, Prob, nMaxLen+1, nMaxLen+1);
			//ת
			//ӵǰλ+-1λת
			lout_assert(nMaxLen > 1);
			for (int i=1; i<=nMaxLen; i++) {
				//±1ʼ
				memset(m_pLenJumpProb[i], 0, (nMaxLen+1)*sizeof(m_pLenJumpProb[i][0]) );
				if ( i== 1 ) {
					m_pLenJumpProb[i][i+1] = 1;
				} else if ( i == nMaxLen ) {
					m_pLenJumpProb[i][i-1] = 1;
				} else {
					m_pLenJumpProb[i][i+1] = 0.5;
					m_pLenJumpProb[i][i-1] = 0.5;
				}
			}

			//ӵǰλתÿλ
// 			for (int i=1; i<=nMaxLen; i++) {
// 				m_pLenJumpProb[i][0] = 0;
// 				for (int j=1; j<=nMaxLen; j++)
// 					m_pLenJumpProb[i][j] = 1.0 / nMaxLen;
// 			}
		}
		/// 
		/**
		 * \param [in] pSeq ʼУͬʱزСڲı䳤ȣҪpSeqĿռ㹻
		 * \param [in] nLen гʼ, ı
		 * \param [in] nTimes 1Ϊһ
		 */
		virtual void Sampling(VocabID *pSeq, int &nLen, int nTimes = 1);
		/// һβ, ıкͳ
		virtual void OneSampling(VocabID *pSeq, int &nLen);
		/// ȼСʱg( x{nLen:nLen+nDim-1} | x{0:nLen-1} )
		virtual LogP g_LogProb(VocabID *pSeq, int nLen, int nDim);
		/// ʱ,g( x{nLen:nLen+nDim-1} | x{0:nLen-1} )
		virtual LogP g_Sampling(VocabID *pSeq, int nLen, int nDim);

		/// wangbin 2014/8/27 µĲ˷λò롢ɾһı䳤
		virtual void OneSamplingEx(VocabID *pSeq, int &nLen);
		/// pSeqеposλòһԪ
		virtual void g_SeqExtend(VocabID *pNew, VocabID *pSeq, int nLen, int pos, VocabID new_value=0);
		/// pSeqеposλɾһԪ
		virtual void g_SeqCut(VocabID *pNew, VocabID *pSeq, int nLen, int pos);
		/// OneSamplingExOneSamplingExãԵ (nLenбnew_posλ)
		virtual LogP g_MarginalProb(VocabID *pSeq, int nLen, int new_pos);
		/// posλýв
		virtual LogP GibbsSample(VocabID *pSeq, int nLen, int pos, bool bSample = true);
		/// λһGibbs
		virtual void GibbsSample(VocabID *pSeq, int nLen);

		/// wangbin 2014/8/27 ·2
		virtual void OneSamplingEx2(VocabID *pSeq, int &nLen);


		/// ü㸽Ȩغͣڲʱڱ䶯Ľǵǰλõֵ˲Ҫ仰ĸ
//		virtual LogP GetWeightSum(VocabID *pSeq, int nLen, int nPos);

		/// ȷһtools_modelеһģ
		virtual void ExactNormalization();
		/// ƹһ
		virtual void ApproxNormalization(int nSampleNum = 10000, int nHotNum = 3000, int nStep = 1);

		virtual void ReadT(wbFile &file)
		{
			SAFE_DELETE_DARRAY(m_pLenJumpProb, m_nMaxLen);
			MLmodel::ReadT(file);
			SetMaxLength(m_nMaxLen); //ʼ
		}
	};


	/** 
	 * \class
	 * \brief CDƵJRFģͣʹýg(x|y)
	 */
	class CDmodel_len : public CDmodel
	{
	public:
		wbArray<LogP> m_aApproxLogNormFactors; ///< ƵĹһϵ

		wbArray<LogP> m_aAvgLogNormFactors; ///< zetaʱҪƽĹһϵ

		//int t;
		//int t0;
		double pi_min;
		double pi_max; ///< ĳȸpi
		int j_max; ///< ĳj
		int j_zero; ///< ¼Ϊ0zetajڹƵԵzetaҪһzetaΪ0

		wbArray<double> m_aAverageNum; ///< ÿȵƽִ, ÿmin-batchĳȷֲ
		int m_nLocalTimes; ///< ֲһθ²Ĳ

		wbFile fileDbg;
		wbArray<double> m_aSampleLenProb; ///< ¼еĲĳȷֲ, debug
		int m_nSampleTotalNum; ///< ¼Ŀ, debug

		bool bInitModel; ///< Ƿʼzeta ǸʼģͣΪfalseΪtrue

	public:
		CDmodel_len(Vocab *vocab, int maxLen = 10): CDmodel(vocab, maxLen) {
// 			t = 0;
// 			t0 = 2e6;
			pi_min = 0;
			pi_max = 0;
			j_max = 0;
			j_zero = 0;
			m_nLocalTimes = 0;
			bInitModel = true;

			SetMaxLength_len(maxLen);

			//fileDbg.Open("SAMS.dbg", "wt");
			m_nSampleTotalNum = 0;
			m_aSampleLenProb.SetNum(c_nMaxLen);
			m_aSampleLenProb.Fill(0);
			
		};

		void SetMaxLength_len(int nMaxLen) {
			//CDmodel::SetMaxLength(nMaxLen);
			m_nMaxLen = nMaxLen;
			m_aAverageNum.SetNum(nMaxLen+1);
			m_aAverageNum.Fill(0);
			m_aApproxLogNormFactors.SetNum(nMaxLen+1);
			m_aAvgLogNormFactors.SetNum(nMaxLen+1);
			for (int i=0; i<=nMaxLen; i++) {
				m_aApproxLogNormFactors[i] = (i-1) * Prob2LogP(m_pVocab->nSize); //ȷֲֵ
				m_aAvgLogNormFactors[i] = 0;
			}
		}

		virtual void SetMaxLength(int nMaxLen)
		{
			CDmodel::SetMaxLength(nMaxLen);
			SetMaxLength_len(nMaxLen);
		}

		virtual LogP NormalizeProb(LogP logP, int nLen) { 
			//һ
#ifndef NotEstimateZeta
			return logP - (m_aLogNormFactors[1] - m_aApproxLogNormFactors[1] + m_aApproxLogNormFactors[nLen]) + Prob2LogP(m_aLenProb[nLen]); 
#else
			return logP - m_aLogNormFactors[nLen] + Prob2LogP(m_aLenProb[nLen]); //ùƵzeta
#endif
		}

		//ͬʱ¹һrΪ
		void SamplingAndEstimation(VocabID *pSeq, int &nLen, double r, bool bOut = true)
		{
			//ڵ̲߳
			Sampling(pSeq, nLen, 1);

			//ͳƲÿȵ
			//SAͳƣΪ˷ֹ̳߳ͻĶڲ
			m_nLocalTimes++; //¼
			for (int i=1; i<=m_nMaxLen; i++) {

				if ( i!= nLen ) {
					m_aAverageNum[i] = m_aAverageNum[i] - m_aAverageNum[i] / m_nLocalTimes;
				} else {
					m_aAverageNum[i] = m_aAverageNum[i] + (1.0 - m_aAverageNum[i]) / m_nLocalTimes;
				}

			}

			UpdateApproximateValues(r, -1, bOut);
		}

		//dbgϢ
		void OutDbg()
		{
			if (!fileDbg.Good())
				return;

			fileDbg.Print("\n");
			for (int i=1; i<=m_nMaxLen; i++)
				fileDbg.Print("%lf\t", m_aLenProb[i]); //ӡ׼ȷĳȷֲ
			fileDbg.Print("\n");
			for (int i=1; i<=m_nMaxLen; i++)
				fileDbg.Print("%lf\t", m_aSampleLenProb[i]); //ӡȷֲ
			fileDbg.Print("\n");

			for (int i=1; i<=m_nMaxLen; i++)
				fileDbg.Print("%lf\t", m_aAverageNum[i]); //ֳȷֲ
			fileDbg.Print("\n");

			for (int i=1; i<=m_nMaxLen; i++)
				fileDbg.Print("%lf\t", m_aLogNormFactors[i]);
			fileDbg.Print("\n");
			for (int i=1; i<=m_nMaxLen; i++)
				fileDbg.Print("%lf\t", m_aLogNormFactors[1] - m_aApproxLogNormFactors[1] + m_aApproxLogNormFactors[i] );
			fileDbg.Print("\n");
			for (int i=1; i<=m_nMaxLen; i++)
				fileDbg.Print("%lf\t", m_aLogNormFactors[1] - m_aAvgLogNormFactors[1] + m_aAvgLogNormFactors[i] );
			fileDbg.Print("\n");
		}

		//zetaй淶ǵj_zeroΪ0
		void NormalZeta(LogP *pZeta)
		{
			double d = pZeta[j_zero];
			for (int j=1; j<=m_nMaxLen; j++)
				pZeta[j] -= d;
		}
		//趨ȷֲ
		void SetLenProb(Prob *pLenProb = NULL)
		{
			if (pLenProb)
			{
				for (int i=1; i<=m_nMaxLen; i++) {
					m_aLenProb[i] = pLenProb[i];
				}
			}

			lout<<"Len Prob (Before smooth): "<<endl;
			for (int i=1; i<=m_nMaxLen; i++) {
				lout<<m_aLenProb[i]<<"\t";
			}
			lout<<endl;
			
			//ԳȽм򵥵ƽֹΪ0ĳ
			bool bNeedSmooth = false;
			for (int i=1; i<=m_nMaxLen; i++) {
				if (m_aLenProb[i] == 0)  {
					bNeedSmooth = true;
					break;
				}
			}

			if (bNeedSmooth) {
				lout<<"Need LenProb Smooth..."<<endl;
				for (int i=1; i<=m_nMaxLen; i++) {
					if ( m_aLenProb[i] < 1e-5 ) {
						lout<<"len="<<i<<" "<<m_aLenProb[i]<<" => "<<1e-5<<endl;
						m_aLenProb[i] = 1e-5;
					}
				}
				LineNormalize(m_aLenProb+1, m_nMaxLen);

				lout<<"Len Prob (after smooth): "<<endl;
				for (int i=1; i<=m_nMaxLen; i++) {
					lout<<m_aLenProb[i]<<"\t";
				}
				lout<<endl;
			}
			
			//ȷpi_min
			pi_min = 1;
			for (int i=1; i<=m_nMaxLen; i++) {
				pi_min = min(pi_min, m_aLenProb[i]);
			}


			if ( pi_min <= 0 ) {
				lout_error("pi_min["<<pi_min<<"] <= 0");
			}
			lout_variable(pi_min);

			pi_max = m_aLenProb[1];
			j_max = 1;
			for (int j=2; j<=m_nMaxLen; j++) {
				if ( m_aLenProb[j] > pi_max ) {
					pi_max = m_aLenProb[j];
					j_max = j;
				}
			}
			lout_variable(pi_max);
			lout_variable(j_max);

			j_zero = j_max; ///< Ϊj_max
			lout_variable(j_zero);
		}
		//ʼ,ڲ֮ǰ
		void InitSampling()
		{
			SetLenProb();


			//ֵ
			for (int i=1; i<=m_nMaxLen; i++) {
				//Ǵļжģ͵Ļm_aApproxLogNormFactorsͲܱ¸ֵ
				if (bInitModel) {
					m_aApproxLogNormFactors[i] = i * Prob2LogP(m_pVocab->nSize); //ȷֲֵ
				}
				m_aAvgLogNormFactors[i] = 0;
			}
			NormalZeta(m_aApproxLogNormFactors);
			//NormalZeta(m_aAvgLogNormFactors);
		}

		//
		virtual void OneSampling(VocabID *pSeq, int &nLen)
		{
// 			if (pi_min <= 0) {  //һβʼ
// 				InitSampling();
// 				OutDbg();
// 			}

			//
			CDmodel::OneSampling(pSeq, nLen);

			//ͳƲÿȵ
			//SAͳƣΪ˷ֹ̳߳ͻĶڲ
// 			m_nLocalTimes++; //¼
// 			for (int i=1; i<=m_nMaxLen; i++) {
// 
// 				if ( i!= nLen ) {
// 					m_aAverageNum[i] = m_aAverageNum[i] - m_aAverageNum[i] / m_nLocalTimes;
// 				} else {
// 					m_aAverageNum[i] = m_aAverageNum[i] + (1.0 - m_aAverageNum[i]) / m_nLocalTimes;
// 				}
// 
// 			}

			
		}

		/// ¹һϵ,rΪѧϰ, t>0ȡƽƽtrue-pipi֮
		double UpdateApproximateValues(double r, int t=-1, bool bOutDbg = false)
		{

			//r = min(pi_min, r); 

			//ͳÿÿȵķֲ
			m_nSampleTotalNum += m_nLocalTimes;
			for (int i=1; i<=m_nMaxLen; i++) {
				m_aSampleLenProb[i] += (m_aAverageNum[i]-m_aSampleLenProb[i]) * m_nLocalTimes/m_nSampleTotalNum;
			}

			// \zeta 
			for (int i=1; i<=m_nMaxLen; i++) {
#ifdef _Norm
				//΢ĵzeta
				m_aApproxLogNormFactors[i] += r /*min(r, m_aLenProb[i]) */ * ( m_aAverageNum[i] / m_aLenProb[i] );
#else
				m_aApproxLogNormFactors[i] += r /*min(r, m_aLenProb[i]) */* ( m_aAverageNum[i] / m_aLenProb[i] );
#endif
			}
			NormalZeta(m_aApproxLogNormFactors);

			if ( t>1 ) //ȡƽ
			{
// 				lout<<"[ ============ avg zeta ============]"<<endl;
// 				lout<<m_aAverageNum[1]<<endl;
// 				lout<<m_aApproxLogNormFactors[1]<<endl;
				for (int i=1; i<=m_nMaxLen; i++) {
					m_aAvgLogNormFactors[i] += 1.0 / t * ( m_aApproxLogNormFactors[i] - m_aAvgLogNormFactors[i] );
				}
			} else {
				for (int i=1; i<=m_nMaxLen; i++) {
					m_aAvgLogNormFactors[i] = m_aApproxLogNormFactors[i];
				}
			}
			
			if (bOutDbg)
				OutDbg(); //ԿƴӡƵ

			//¼¼ƽ
			m_nLocalTimes = 0;
			m_aAverageNum.Fill(0);

			//жpipiǷѾ㹻ӽ
			double dDiffRate = 0;
			for (int i=1; i<=m_nMaxLen; i++) {
				dDiffRate = max( dDiffRate, fabs(m_aSampleLenProb[i]-m_aLenProb[i])/m_aLenProb[i] );
			}
			return dDiffRate;

		}

		/// aAvgLogNormFactor в aApproxLogNormFactor в
		/* ʹAvgZetaȻֵעʹúҪȥ!! */
		void ExchangeAvgNoavgZeta()
		{
			double d;
			for (int i=1; i<=m_nMaxLen; i++) {
				d = m_aAvgLogNormFactors[i];
				m_aAvgLogNormFactors[i] = m_aApproxLogNormFactors[i];
				m_aApproxLogNormFactors[i] = d;
			}
		}

		virtual void WriteT(wbFile &file)
		{
			CDmodel::WriteT(file);
			for (int i=1; i<=m_nMaxLen; i++) {
				file.Print("[%d]\t%lf %lf\n", i, m_aLenProb[i], m_aApproxLogNormFactors[i]);
			}
		}
		virtual void ReadT(wbFile &file) 
		{
			CDmodel::ReadT(file);
			SetMaxLength_len(m_nMaxLen);
			m_aLogNormFactors[1] = Normalization(1);

			//жǷΪ֮ڳȷֲzeta
			char *pLine = NULL;
			int nLen = 0;
			for (int i=1; i<=m_nMaxLen; i++) {
				pLine = file.GetLine();
				if ( pLine == NULL ) {
					lout_warning("init-modelpizetaʼzeta=0");
					//bInitModel = true;
					m_aApproxLogNormFactors.SetNum(m_nMaxLen+1);
					m_aApproxLogNormFactors.Fill(0);
					break;
				}
				sscanf(pLine, "[%d]\t%lf %lf\n", &nLen, &m_aLenProb[i], &m_aApproxLogNormFactors[i]);
			}

			
// 			for (int i=1; i<=m_nMaxLen; i++) {
// 				fscanf(file, "[%d]\t%lf %lf\n", &nLen, &m_aLenProb[i], &m_aApproxLogNormFactors[i]);
// 			}
			SetLenProb();
			bInitModel = false;
		}
		virtual void ExactNormalization()
		{
			Normalization(true);
			for (int i=1; i<=m_nMaxLen; i++)
				m_aApproxLogNormFactors[i] = m_aLogNormFactors[i] - m_aLogNormFactors[j_zero];
		}
		void ApproxNormalization( VocabID *pSeq, int &nSeq, double threshold = 0.1, int nMax = 0, const char *txt = NULL)
		{
			wbFile file("ApproxNormalization.dbg", "wt");

			wbArray<int> aSampleNum(m_nMaxLen+1);
			aSampleNum.Fill(0); //¼ÿȵĳִ
			int nTime;

			titlePrecent(0, true, nMax, "");
			for (nTime = 1; nTime != nMax; nTime++)
			{
				OneSampling(pSeq, nSeq);
				aSampleNum[nSeq]++;

				if ( nTime % 1000 == 0 ) {
					file.Print("\n");
					for (int i=1; i<=m_nMaxLen; i++)
						file.Print("%lf\t", m_aLenProb[i]); //ӡ׼ȷĳȷֲ
					file.Print("\n");
					for (int i=1; i<=m_nMaxLen; i++)
						file.Print("%lf\t", 1.0*aSampleNum[i]/nTime); //ӡȷֲ
					file.Print("\n");
				}
				

				//if ( nTime > 3000 ) {
					double r = min( m_aLenProb[nSeq], 1.0 / nTime );
					m_aApproxLogNormFactors[nSeq] += r * (1.0 / m_aLenProb[nSeq]);
					NormalZeta(m_aApproxLogNormFactors);
				//}
				

				double tol = 0;
				for (int j=1; j<=m_nMaxLen; j++) {
// 					d1 += 1.0 * aSampleNum[j] / nTime * m_aLenProb[j];
// 					d2 += pow( 1.0 * aSampleNum[j] / nTime, 2 );
// 					d3 += pow( m_aLenProb[j], 2 );
					double d = fabs( 1.0 * aSampleNum[j] / nTime - m_aLenProb[j] ) / m_aLenProb[j];
					tol = max( tol, d );
				}

				if ( nTime % 100 == 0 ) {
					lout<<"t="<<nTime<<" tol="<<tol<<endl;
					double ppl, likelihood;
					if ( txt ) {
						this->PPL(txt, ppl, likelihood);
						lout<<"ppl="<<ppl<<" likelihood="<<likelihood<<endl;
					}
				}

				if ( tol < threshold )
					break;

				titlePrecent(nTime);
			}

			lout<<"ApproxNormalization: tmax="<<nTime<<endl;
		}
	};
}