// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// TDRF models, based class
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


/** 
 * \file
 * \author wangbin
 * \date 2014-03-3
 * \brief JRFģ, MLѵ˰ǰ㷨
 */
#pragma once
#include "wbJRFmodel.h"

namespace wbJRF
{
	const int c_nMaxOrder = ngram_maxOrder;       ///< 
	const int c_nMaxLen = 1000;		///< 󳤶

	const int c_nMaxFindFeatures = 1024;  ///< FindFeatrue ص

	/** 
	 * \class
	 * \brief 
	 */
	class VecIter
	{
	public:
		int *m_pBuf; ///< ռ
		int m_nDim; ///< ά
		int m_nMin; ///< ÿάȵСֵ
		int m_nMax; ///< ÿάȵֵ
		VecIter(int *pBuf, int nDim, int nMin, int nMax):m_pBuf(pBuf),m_nDim(nDim), m_nMin(nMin), m_nMax(nMax) {
			Init();
		}
		void Init() {
			for (int i=0; i<m_nDim; i++)
				m_pBuf[i] = m_nMin;
		}
		bool Next() {
			m_pBuf[m_nDim-1]++;
			for (int i=m_nDim-1; i>0; i--) {
				if (m_pBuf[i] > m_nMax) {
					m_pBuf[i-1]++;
					m_pBuf[i] = m_nMin;
				} else {
					break;
				}
			}
		
			return m_pBuf[0]<=m_nMax;
		}
		bool bGood() { return m_nDim>0 && m_pBuf[0]<=m_nMax; }
	};

	/** 
	 * \class
	 * \brief , Կÿάȵĵ
	 */
	class DimIter
	{
	public:
		int *m_pBuf; ///< ռ
		int m_nDim; ///< ά
		wbArray<wbArray<int>*> m_aDimArray; ///< ¼ÿάȵĵ
		wbArray<int> m_aDimCur; ///< ¼ÿάȵĵǰλ
		DimIter(int *pBuf, int nDim):m_pBuf(pBuf), m_nDim(nDim) {};
		void Init()
		{
			for (int i=0; i<m_nDim; i++)
				m_aDimCur[i] = 0;
			for (int i=0; i<m_nDim; i++)
				m_pBuf[i] = m_aDimArray[i]->Get(m_aDimCur[i]);
		}
		bool Next()
		{
			m_aDimCur[m_nDim-1]++;
			for (int i=m_nDim-1; i>0; i--) {
				if (m_aDimCur[i] >= m_aDimArray[i]->GetNum()) {
					m_aDimCur[i-1]++;
					m_aDimCur[i] = 0;
				} else {
					break;
				}
			}

			if (m_aDimCur[0] < m_aDimArray[0]->GetNum()) {
				for (int i=0; i<m_nDim; i++)
					m_pBuf[i] = m_aDimArray[i]->Get(m_aDimCur[i]);
				return true;
			} else {
				return false;
			}
		}
		bool bGood() { return m_aDimCur[0] < m_aDimArray[0]->GetNum(); }
	};

	/** 
	 * \class 
	 * \brief 汾ָļά
	 */
	class SupIter
	{
	public:
		int *m_pBuf; ///< ռ
		int m_nDim; ///< ά
		int m_nMin; ///< ÿάȵСֵȡ
		int m_nMax; ///< ÿάȵֵȡ
		const int *m_pFlag; ///< һάҪ
		bool m_bEnd; ///< Ƿ
		/// flagЩάҪҪάȣflagֵ<0ֱʹJumpGramģNULL򲻼
		SupIter(int *pBuf, int nDim, int nMin, int nMax, const int *flag = NULL)
		{
			m_pBuf = pBuf;
			m_nDim = nDim;
			m_nMin = nMin;
			m_nMax = nMax;
			m_pFlag = flag;
			Init();
		}
		void Init() {
			for (int i=0; i<m_nDim; i++) {
				if (m_pFlag[i] < 0)
					m_pBuf[i] = m_nMin;
				else
					m_pBuf[i] = m_pFlag[i];
			}
			m_bEnd = false;
		}
		bool Next() {

			int inc;
			
			for (inc=m_nDim-1; 
				m_pFlag[inc]>=0 && inc>=0; 
				inc--) {} ///ҵһά
			
			

			if (inc < 0) {
				m_bEnd = true;
				return false;
			}

			m_pBuf[inc]++; //1

			while(m_pBuf[inc] > m_nMax) {

				m_pBuf[inc] = m_nMin;
				for (inc--;m_pFlag[inc]>=0 && inc>=0; 
					inc--) {} ///ҵһά

				if (inc < 0) {
					m_bEnd = true;
					return false;
				}
				m_pBuf[inc]++;
			}

			return true;
		}
		bool bGood() { return !m_bEnd; }
	};

	/** 
	 * \class
	 * \brief Ϣ
	 */
	class Msg
	{
	public:
		double *pBuf;
		int nSize;  ///< ÿһάȵSize
		int nBufSize; ///< ٵĿռС
	public:
		Msg(int nMsgOrder/*ϢһΪģͽ-1*/, int nVocabSize):nSize(nVocabSize)
		{
			nBufSize = 1;
			//Ϣ
			for (int i=0; i<nMsgOrder; i++)
				nBufSize *= nVocabSize; 
			pBuf = new double[nBufSize];
	
			Fill(0);
		}
		~Msg()
		{
			SAFE_DELETE_ARRAY(pBuf);
		}
		void Fill(double d)
		{
			for (int i=0; i<nBufSize; i++)
				pBuf[i] = 0;
		}
		/// һά
		double& Get(int i) { return pBuf[i]; }
		/// ά
		double& Get(int i, int j) { return pBuf[i*nSize+j]; }
		/// 
		double& Get(int *pSeq, int nDim)
		{
			int nIndex = pSeq[0];
			for (int i=0; i<nDim-1; i++)
			{
				nIndex = nIndex * nSize + pSeq[i+1];
			}
			return pBuf[nIndex];
		}
	};

	struct GramPack
	{
		int nFeatId;	///< ngramid
		short nPos;   ///< λãһclusterдŵ֮λõҪ¼λ
	};

	/** 
	 * \class
	 * \brief MLƵJRFģͣǰ㷨Լһϵ
	 */
	class MLmodel : public model
	{
	public:
		short m_nOrder; ///< ģͽ, ML˵֧FeatureΪSuperGramNgramJumpGram,Ҫ֧

		wbArray<Msg*> m_aAlpha; ///< ǰϢ
		wbArray<Msg*> m_aBeta;  ///< Ϣ
		//wbArray<wbArray<GramPack>> m_aClusters; ///< ÿλõcluster,ÿλü

		int m_nClusterNum; ///< ¼ǰǰ㷨ҪclusterĿ

		wbTrie<VocabID, wbArray<int>*> m_aGramIndex; ///<ͨGramпֻһ 
		wbArray<wbArray<int>*> m_aGramPack; ///< m_aGramIndexеָ룬ͷڴ
		/**
			ͳһtokenwordclassidӳ䵽ͳһռ䣬£
			0token_num-1ΪtokenӦtokenӦidͬ
			token_numtoken_num+word_num-1ΪwordӦindex = word_id + token_num
			֮classӦindex = class_id + token_num + word_num
		*/
		int m_nWordIndexOffset; ///< ڽword-idӳ䵽ϣindexoffset
		int m_nClassIndexOffset; ///< class-idӳ䵽ϣindexoffsetclass-id+offsetȻϷ
		
		double *m_pdFeatVar;		///< ׼, Ϊ˶normalization
		double *m_pdFeatMean;	///< ֵΪ˶normalization

		wbArray<LogP> m_aLogNormFactors; ///< logһϵ
		wbArray<double> m_aLenProb; ///< ȵֲ, ܻõ

		///< ±ŻңԲؿǴֵļʱԶ
		bool m_bContainJumpNgram; ///< ¼ǷҪȥJumpNgram
		bool m_bContainTokenGram; ///< ¼ǷTokengram
		bool m_bContainClassGram;
		bool m_bContainClassGram_jump;
		bool m_bContainWordGram;
		bool m_bContainWordGram_jump; ///< wordgramҴjumpϢ
		bool m_bContainPredictGram;
		bool m_bContainMixGram;
		bool m_bContainMixGram_cpw; ///< f(c,c,c,w)
		bool m_bContainMixGram_cpw_jump; ///< f(c, -1, w)
		bool m_bContainMixGram_wpc; ///< f(w,w,c);

		/*Ҫ֤full class/word ngram ÿλöЧjumpֵ*/
		/* 
			findFeatureУĿǰҪҵorderڵڴֵ֮Ѱģʽjump(w,-1,-1,-1,-1,w)
			ģʽԡ
			Ҫ֤ģе.
		*/
		int m_nFullGramLimitedOrder;  ///< Ŀǰorderڴֵʱֻضjump
		int m_nFullGramMaxOrder; ///< fullgram

	public:
		MLmodel(Vocab *vocab, int maxLen) : model(vocab, maxLen), m_nOrder(-1)
		{
			//һϵ
			for (int i=0; i<=m_nMaxLen; i++)
				m_aLogNormFactors[i] = 0;

// 			m_aLenProb.SetNum(maxLen+1);
// 			m_aLenProb.Fill(1);

			//Ϣռ
			//Reset(order, maxLen);
			m_bContainJumpNgram = false;
			m_bContainTokenGram = false;
			m_bContainClassGram = false;
			m_bContainClassGram_jump = false;
			m_bContainWordGram = false;
			m_bContainWordGram_jump = false;
			m_bContainPredictGram = false;
			m_bContainMixGram = false;
			m_bContainMixGram_cpw = false;
			m_bContainMixGram_cpw_jump = false;
			m_bContainMixGram_wpc = false;

			m_nFullGramLimitedOrder = 5;

			m_nWordIndexOffset = m_pVocab->aVocabInfo.GetNum();
			m_nClassIndexOffset = m_nWordIndexOffset + m_pVocab->nWordIdTop;
			
			m_pdFeatVar = NULL;
			m_pdFeatMean = NULL;
		}
		~MLmodel()
		{
			//Msg
			for (int i=0; i<m_aAlpha.GetNum(); i++)
				SAFE_DELETE(m_aAlpha[i]);
			for (int i=0; i<m_aBeta.GetNum(); i++)
				SAFE_DELETE(m_aBeta[i]);

			SAFE_DELETE_ARRAY(m_pdFeatVar);
			SAFE_DELETE_ARRAY(m_pdFeatMean);
		}
		/// д
// 		void LoadInitFeatureOrModel(wbFile &file, bool bFeatures = true, bool bAllockMsgBuffer = true)
// 		{
// 			if ()
// 			model::ReadFeatures(file);
// 			lout_variable(m_nFeatNum)
// 			
// 			
// 		}

		/// ʼģͻĳʼ
		virtual void PrepareInitModel(bool bAllockMsgBuffer = true)
		{
			lout_variable(m_nFeatNum);
			lout_variable(m_nMaxLen);

			//Ƿ, ͬʱΪm_nOrderֵ
			m_bContainTokenGram = false;
			m_bContainJumpNgram = false;
			m_bContainClassGram = false;
			m_bContainWordGram = false;
			m_bContainWordGram_jump = false;
			m_bContainPredictGram = false;
			m_bContainMixGram = false;
			m_bContainMixGram_cpw = false;
			m_bContainMixGram_cpw_jump = false;
			m_bContainMixGram_wpc = false;
			m_nFullGramMaxOrder = 0;
			m_nOrder = 0;
			for (int i=0; i<m_nFeatNum; i++) {
				if (m_aFeatures[i]->type == base) {
					lout_error("Ͳngram: "<<m_aFeatures[i]->type);
				}
				if (m_aFeatures[i]->type <= super)
					m_bContainTokenGram = true;
				if (m_aFeatures[i]->type == jgram)
					m_bContainJumpNgram = true;
				if (m_aFeatures[i]->type == cgram) {
					m_bContainClassGram = true;
					if ( m_aFeatures[i]->bJump() ) 
						m_bContainClassGram_jump = true;
				}
				if (m_aFeatures[i]->type == wgram) {
					m_bContainWordGram = true;
					if ( m_aFeatures[i]->bJump() ) m_bContainWordGram_jump = true;
				}
				if (m_aFeatures[i]->type == pgram)
					m_bContainPredictGram = true;
				if (m_aFeatures[i]->type == ggram) {
					m_bContainMixGram = true;
					GlobalNgram *pGlobal = (GlobalNgram*)m_aFeatures[i];
					if ( true == ( m_bContainMixGram_cpw = pGlobal->Is_cpw() ) )
						m_bContainMixGram_cpw_jump = pGlobal->bJump();
					m_bContainMixGram_wpc = pGlobal->Is_wpc();
					//lout_error("MLmode cannot support ggram type feature!");
				}

				Ngram *pNgram = (Ngram*)m_aFeatures[i];
				int nFeatOrder = ((Ngram*)m_aFeatures[i])->order;
				if ( nFeatOrder >= m_nFullGramLimitedOrder ) {
					if ( pNgram->bSkipBigram() == false )
						lout_error("Model: Do not support the non-skip-bigram for order "<<nFeatOrder);
				}
				if ( pNgram->bJump() == false )
					m_nFullGramMaxOrder = max(m_nFullGramMaxOrder, nFeatOrder);

				m_nOrder = max(m_nOrder, ((Ngram*)m_aFeatures[i])->order);
			}
			if (m_bContainClassGram || m_bContainWordGram || m_bContainMixGram) {
				if (m_pVocab->aVocabInfo.GetNum() == 0) {
					lout_error("classϢtag-vocabļ");
				}
			}

			lout_variable(m_bContainTokenGram);
			lout_variable(m_bContainJumpNgram);
			lout_variable(m_bContainClassGram);
			lout_variable(m_bContainClassGram_jump);
			lout_variable(m_bContainWordGram);
			lout_variable(m_bContainWordGram_jump);
			lout_variable(m_bContainPredictGram);
			lout_variable(m_bContainMixGram);
			lout_variable(m_bContainMixGram_cpw);
			lout_variable(m_bContainMixGram_cpw_jump);
			lout_variable(m_bContainMixGram_wpc);
			lout_variable(m_nFullGramLimitedOrder);
			lout_variable(m_nFullGramMaxOrder);

			FeaturesReHash(); //

			if (bAllockMsgBuffer)
				Reset(m_nOrder, m_nMaxLen);

			
			m_aLogNormFactors[1] = Normalization(1);

			lout_variable(m_nOrder);
		}

		/// 㺯GramGram־ӳ䵽ϣеindex
		void GetGramIndex(Ngram *pGram, VocabID *aIndex);

		/// ͬʱ
		//virtual void AddFeature(Feature *pf, double dValue  = 0 );
		/// ͬʱ
		virtual void CleanFeatures();
		/// ¸
		void FeaturesReHash();
		/// еĸ, bNorm:Ƿһ
		virtual LogP GetLogProb(VocabID *pSeq, int nLen, bool bNorm = true);
		/// ü㸽Ȩغͣڲʱڱ䶯Ľǵǰλõֵ˲Ҫ仰ĸ
		virtual LogP GetWeightSum(VocabID *pSeq, int nLen, int nPos);
		/// ȡ
		virtual void ExtractFeatures(VocabID *pSeq, int nLen, int *pVec);
		/// Gram
		/** 
		 * \param [in] pSeq 
		 * \param [in] nLen гȣɸѡsuperNgram
		 * \param [in] h GramλãɸѡsuperNgram
		 * \param [in] nOrder Gram
		 * \param [out] a idļ
		 */
		void FindFeatures(wbArray<int> &a, VocabID *pSeq, int nLen, int h, int nOrder );

		/// FindFeaturesеãjump
		void FindJumpFeatures(wbArray<wbArray<int>**> &aFindList, VocabID *aIndex, int nOrder);
		/// Gram
		/** 
		 * ЧͬһFindFeatureֻʹָ뷵ضwbArray<int> ,Ҫ֤pFeat㹻Ŀռ
		 * \param [out] pFeat idļ
		 * \param [out] nFeat idĸע⣬ҪⲿʼڲʼnFeatʹóԷFindFeatrue׷
		 * \param [in] nMax pFeatٵĿռ
		 */
		//void FindFeatures(int *pFeat, int &nFeat, int nMax, VocabID *pSeq, int nLen, int h, int nOrder);
		/// һгֵ
		/** 
		 * \param [in] pSeq 
		 * \param [in] nLen гȣɸѡsuperNgram
		 * \param [out] a idļ
		 */
		void FindFeatures(wbArray<int> &a, VocabID *pSeq, int nLen);

		/// ´ģͣҪؽϢ
		void Reset(int nOrder, int nMaxLen);

		/// ǰCluster
		void AllocateFeatrues(int nLen);
		/// hclusterенpSeqValueͣ pSeqΪУΪnLen
		LogP ClusterSum(VocabID *pSeq, int nLen, int h);
		/// ǰ㷨
		void Forward(int nLen);
		/// 㷨
		void Backward(int nLen);
		/// ǰ㷨Ϣ
		void ForwardBackward(int nLen){
			AllocateFeatrues(nLen);
			Forward(nLen);
			Backward(nLen);
		}

		/// ԳnLenģ͹һ
		LogP Normalization(int nLen);
		/// ģͽйһ; bPrint Ƿӡһ
		void Normalization(bool bPrint = false);
		/// Ըʹһ, nLenΪǰ
		virtual LogP NormalizeProb(LogP logP, int nLen) { return logP - m_logNormalizationFactor; }

		/// ԵpGramϵĸ
		LogP MarginalizeLogProb(int nLen, int h, VocabID *pGram, int nDim);

		/// 㳤ΪnLenģ͵p[fi], ʹǰҪȵǰ㷨
		LogP LogExpectation(int nLen, int i);
		/// 㳤ΪnLenģ͵p[fi] i = 1,...,F
		void LogExpectation(int nLen, LogP *pdExp);
// 		//㳤ΪnLenģ͵p[fi^2], ʹǰҪȵǰ㷨
// 				LogP LogExpSqu(int nLen, int i);
		/// 
		virtual void LogExpectation(LogP *pdExp);

		/// òıҪһιһ
		virtual void SetParams(double *pdParams, bool bNormal = true)
		{
			model::SetParams(pdParams);
#ifdef _FeatNorm
			if ( m_pdFeatVar ) {
				for (int i=0; i<m_nFeatNum; i++)
					m_aValues[i] /= m_pdFeatVar[i];
			}
#endif
			if (bNormal)
				Normalization();
			else
				m_aLogNormFactors[1] = Normalization(1); //ֻ㳤Ϊ1Ĺһϵ
		}
		/// ȡ
		virtual void GetParams(double *pdParams)
		{
			memcpy(pdParams, m_aValues, sizeof(double) * m_nFeatNum);
#ifdef _FeatNorm
			if ( m_pdFeatVar ){
				for (int i=0; i<m_nFeatNum; i++)
					pdParams[i] *= m_pdFeatVar[i];
			}
#endif
		}
		/// 趨ֵͷҪΪʹ
		virtual void SetFeatMeanVar( const LogP *pdMean, const LogP *pdVar )
		{
#ifdef _FeatNorm
			if (pdMean) {
				SAFE_NEW_ARRAY(m_pdFeatMean, double, m_nFeatNum);
				for (int i=0; i<m_nFeatNum; i++)
					m_pdFeatMean[i] = LogP2Prob(pdMean[i]);
			}
			if ( pdVar ) {
				SAFE_NEW_ARRAY(m_pdFeatVar, double, m_nFeatNum);
				for (int i=0; i<m_nFeatNum; i++)
					m_pdFeatVar[i] = LogP2Prob(pdVar[i]);
			}
#endif
		}
	};

	/** 
	 * \class
	 * \brief ˼·һģͣÿȵһ
	 */
	class MLmodel_len : public MLmodel
	{
	public:
		MLmodel_len(Vocab *vocab, int maxLen = 10): MLmodel(vocab, maxLen) {
			SetLength(maxLen);
		};
		void SetLength(int nLen) {
			m_nMaxLen = nLen;
			m_aLenProb.SetNum(nLen+1);
			m_aLenProb.Fill(1.0/nLen);
		}
		virtual LogP NormalizeProb(LogP logP, int nLen) { 
			//һ
			return logP - m_aLogNormFactors[nLen] + Prob2LogP(m_aLenProb[nLen]); 
		}
		virtual void WriteT(wbFile &file)
		{
			MLmodel::WriteT(file);
			for (int i=1; i<=m_nMaxLen; i++) {
				file.Print("[%d]\t%lf %lf\n", i, m_aLenProb[i], m_aLogNormFactors[i]);
			}
		}
		virtual void ReadT(wbFile &file) 
		{
			MLmodel::ReadT(file);
			SetLength(m_nMaxLen);
			
			int nLen = 0;
			for (int i=1; i<=m_nMaxLen; i++) {
				fscanf(file, "[%d]\t%lf %lf\n", &nLen, &m_aLenProb[i], &m_aLogNormFactors[i]);
			}
		}
	};
}