// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// Pre-defination for TDRF models
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.



/** 
 * \file
 * \author wangbin
 * \date 2014-02-27
 * \brief JRFģ ȫֶ
 */
#pragma once
#include "wbSystem.h"

#define LogP double
#define Prob double
#define INF 1e20
#define Logp_Zero -INF


namespace wbJRF
{
	typedef int VocabID; ///< VocabԪ, ȫididid
	const int VocabID_none = -1; ///< VocabID

	/** 
	 * \struct
	 * \brief VocabInfo,idword-idclass-id
	 */
	struct VocabInfo
	{
		VocabID wid;
		VocabID cid;
	};
	/**
	 * \class
	 * \brief JRFģ͵ֵ䣬Ҫֵ0ʼαţʼ
	 */
	class Vocab
	{
	public:
		int nSize; ///< ֵС
		wbArray<wbString> aWords; ///< ¼ÿʺŵĴַ
		Vocab() {
			//Ĭϳʼɵģ
			nSize = 26;

			char c[2];
			c[1] = '\0';
			for (int i=0; i<26; i++) {
				c[0] = 'a' + i;
				aWords[i] = c;
			}

			nClassIdTop = 0;
			nWordIdTop = 0;
		}
		Vocab(const char *pList) { ///< listӦļÿΪ "id   "
			wbFile file(pList, "rt");
			nSize=0;
			char *pLine;
			while (pLine = file.GetLine()) {
				char *p = strtok(pLine, " \t");
				if (p) {
					if (nSize != atoi(p)) {
						lout_error("ֵ䲻Ǵ0ʼŵģ"<<" nLine = "<<file.nLine);
					}
					aWords[atoi(p)] = strtok(NULL, " \t");
				}
				nSize++;
			}

			nClassIdTop = 0;
			nWordIdTop = 0;
		}
	public: //class֧
		wbArray<VocabInfo> aVocabInfo; ///< ¼global-idword-idӳ
		int nWordIdTop; ///< word id Ͻ+1, 0ţӦĿ
		int nClassIdTop; ///< class id Ͻ+1
		wbArray<wbArray<VocabID>> aWordToTokens; ///< word-idҵwordtoken
		wbArray<wbArray<VocabID>> aClassToTokens; ///< class-idҵclasstoken
		wbTrie<VocabID, VocabID> aWordClassToTokens; ///< wordclassλtoken
		void LoadVocabInfo(const char *pTagVocab) ///< ϢļÿСid  word-id  class-id
		{
			lout<<"Load Vocab Info..."<<endl;
			nWordIdTop = 0;
			nClassIdTop = 0;

			wbFile file(pTagVocab, "rt");
			char *pLine;
			while (pLine = file.GetLine()) {
				char *pId = strtok(pLine, " \t");
				char *pWord = strtok(NULL, " \t");
				char *pClass = strtok(NULL, " \t");
				if (!pId || !pWord || !pClass) { lout_error("line="<<file.nLine<<"; file="<<pTagVocab); }
				int nid = atoi(pId);
				aVocabInfo[nid].wid = atoi(pWord);
				aVocabInfo[nid].cid = atoi(pClass);

				nWordIdTop = max(nWordIdTop, aVocabInfo[nid].wid+1);
				nClassIdTop = max(nClassIdTop, aVocabInfo[nid].cid+1);
			}
			lout_variable(nWordIdTop);
			lout_variable(nClassIdTop);

			CaculateMapping();
		}
		void CaculateMapping() //aWordToTokens  aClassToTokens
		{
			for (int i=0; i<nSize; i++) {
				int nw = aVocabInfo[i].wid;
				int nc = aVocabInfo[i].cid;
				aWordToTokens[nw].Add(i);
				aClassToTokens[nc].Add(i);

				VocabID index[3];
				index[0] = nw;
				index[1] = nc;
				*aWordClassToTokens.Insert(index, 2) = i;
			}
#ifdef _DEBUG
			wbFile file1("vocab1.map", "wt");
			for (int i=0; i<aWordToTokens.GetNum(); i++) {
				file1.Print("%d\t", i);
				for (int n=0; n<aWordToTokens[i].GetNum(); n++)
					file1.Print("%d ", aWordToTokens[i][n]);
				file1.Print("\n");
			}

			wbFile file2("vocab2.map", "wt");
			for (int i=0; i<aClassToTokens.GetNum(); i++) {
				file2.Print("%d\t", i);
				for (int n=0; n<aClassToTokens[i].GetNum(); n++)
					file2.Print("%d ", aClassToTokens[i][n]);
				file2.Print("\n");
			}
#endif
		}
		VocabID FindToken(VocabID w, VocabID c) 
		{ 
			VocabID a[2];
			a[0] = w;
			a[1] = c;
			VocabID *p = aWordClassToTokens.Find(a, 2);
			if (!p)
				return VocabID_none;
			return *p;
		}
	};

	/// Logpprobת
	inline Prob LogP2Prob( LogP x ) {
		return (x<=Logp_Zero/2)? 0:exp((double)(x));
	}
	inline LogP Prob2LogP( Prob x ) {
		return ((x)<=0)? Logp_Zero:log((double)(x));
	}
	///log[exp(x) + exp(y)]
	inline LogP Log_Sum(LogP x, LogP y) {
		return (x>y)? x+Prob2LogP(1+LogP2Prob(y-x)) : y+Prob2LogP(1+LogP2Prob(x-y));
	}

	/// һάֲһ
	/** 
	 * \param [in] pdProbs ķֲ
	 * \param [in] nNum 
	 */
	inline LogP LogLineNormalize(LogP* pdProbs, int nNum)
	{
		LogP dSum = Logp_Zero;
		for (int i=0; i<nNum; i++)
			dSum = Log_Sum(dSum, pdProbs[i]);
		for (int i=0; i<nNum; i++)
			pdProbs[i] -= dSum;
		return dSum;
	}
	inline int LogLineSampling(LogP* pdProbs, int nNum)
	{
		double d = 1.0 * rand() / RAND_MAX;
		int sX = 0;
		double dSum = 0;

		for (sX=0; sX<nNum; sX++) {
			dSum += LogP2Prob(pdProbs[sX]);

			if (fabs(dSum-1) < 1e-5)
				dSum = 1; //ȷ
			if ( dSum == 0 )
				continue; //0

			if (d <= dSum)
				break;
		}
		if (sX >= nNum) {
			cout<<"[LogLineSampling] "<<sX<<"\t"<<dSum<<endl;
			lout_assert(sX<nNum);
		}

		return sX;
	}
	/// һά
	/** 
	 * \param [in] pdProbs УķֲҪһ
	 * \param [in] nNum 
	 * \return õ
	 */
	inline void LineNormalize(Prob* pdProbs, int nNum)
	{
		Prob dSum = 0;
		for (int i=0; i<nNum; i++)
			dSum += pdProbs[i];

		if ( dSum > 0 ) {
			for (int i=0; i<nNum; i++)
				pdProbs[i] /= dSum;
		} else {
			for (int i=0; i<nNum; i++)
				pdProbs[i] = 1.0 / nNum;
		}
		
	}
	inline int LineSampling(Prob* pdProbs, int nNum)
	{
		double d = 1.0 * rand() / RAND_MAX;
		int sX = 0;
		double dSum = 0;

		for (sX=0; sX<nNum; sX++) {
			dSum += pdProbs[sX];

			if (fabs(dSum-1) < 1e-5)
				dSum = 1; //ȷ
			if ( dSum == 0 )
				continue; //0

			if (d <= dSum)
				break;
		}
		if (sX >= nNum) {
			cout<<"[LineSampling] "<<sX<<"\t"<<dSum<<endl;
			lout_assert(sX<nNum);
		}

		return sX;
	}
	/// MHĽ
	inline bool Acceptable(Prob prob)
	{
		double d = 1.0 * rand() / RAND_MAX;
		return d<=prob;
	}
	/// positionһȡn滻ǰnλ
	/** 
	 * \param [in] a Ҫ
	 * \param [in] len 鳤
	 * \param [out] n Ҫĸ
	 */
	inline void RandomPos(int *a, int len, int n)
	{
		if ( n> len ) {
			lout_error("[RandomPos] n("<<n<<") > len("<<len<<") !!");
		}

		for (int i=0; i<n; i++) {
			int s = rand() % (len-i);
			//aеi+sλiλý
			int temp = a[i+s];
			a[i+s] = a[i];
			a[i] = temp;
		}
	}

	/**
	 * \class
	 * \brief ϢΪ˼ȻֵҪÿ Ŀ+Ŀ
	 */
	class CorpusInfo
	{
	public:
		int* m_aFeatCount;  ///< ¼Ŀ
		int* m_aLenCount;   ///< Ŀ
		int m_nTotalWord; ///< ܴʸ
		CorpusInfo():m_aFeatCount(NULL),m_aLenCount(NULL){}
		~CorpusInfo()
		{
			SAFE_DELETE_ARRAY(m_aFeatCount);
			SAFE_DELETE_ARRAY(m_aLenCount);
		}
		void Reset(int nFeatNum, int nMaxLen) 
		{
			SAFE_DELETE_ARRAY(m_aFeatCount);
			SAFE_DELETE_ARRAY(m_aLenCount);
			SAFE_NEW_ARRAY(m_aFeatCount, int, nFeatNum);
			SAFE_NEW_ARRAY(m_aLenCount, int, nMaxLen+1);
			memset(m_aFeatCount, 0, sizeof(int)*nFeatNum);
			memset(m_aLenCount, 0, sizeof(int)*(nMaxLen+1));
			m_nTotalWord = 0;
		}
	};

	/**
	 * \class
	 * \breif ļƣٶȡĳһ
	 */
	class Corpus
	{
	public:
		FILE*	m_pFile; ///< ļ
		int		m_nLine; ///< 
		int		m_nSize; ///< ÿ仰ÿУĳȣΪ˷㴦䣬ÿгһ£е󳤶
		wbArray<wbArray<VocabID>*> aaSeq; ///< ڴ

		//* ޸ĶƸʽڰÿͳһҪܶռ䣬˱һÿ仰Ŀʼλõıٸint
		wbArray<int> m_aSeqBegOffset; ///< ռm_nLine+1һλñǾܳ
		long long m_llSeqOffset; ///< ʼݵλ

		/// corpusϢ
		CorpusInfo m_info;  ///< ˱ⲿֵ

	public:
		Corpus(const char *path = NULL, bool bLoadAll=false):m_pFile(NULL),m_nSize(0),m_nLine(0) { if (path) {Reset(path, bLoadAll);} }
		~Corpus() { 
			Clean();
		}
		void Clean()
		{
			SAFE_FCLOSE(m_pFile); 
			for (int i=0; i<aaSeq.GetNum(); i++)
				SAFE_DELETE(aaSeq[i]);
			aaSeq.Clean();
			m_llSeqOffset = 0;
		}
		void Reset(const char *path, bool bLoadAll=false)
		{
			Clean();
			SAFE_FOPEN(m_pFile, path, "rb");
			fread(&m_nLine, sizeof(int), 1, m_pFile);
			fread(&m_nSize, sizeof(int), 1, m_pFile);
			//¼ÿ仰ʼλõı
			m_aSeqBegOffset.SetNum(m_nLine+1);
			fread(m_aSeqBegOffset.GetBuffer(), sizeof(int), m_nLine+1, m_pFile);
			m_llSeqOffset = _ftelli64(m_pFile);  //濪ʼλ

			if (bLoadAll) // Load all data to memory
			{
				lout<<"[Corpus] Load All Data"<<endl;
				for (int i=0; i<m_nLine; i++) {
					aaSeq.Add() = new wbArray<VocabID>;
					GetLineFromFile(i, *aaSeq[i]);
				}
				lout_assert(aaSeq.GetNum() == m_nLine);
			}

			lout_variable(m_nLine);
			lout_variable(m_nSize);
		}
		/// ļлõnLineеݣ0ʼ
		bool GetLineFromFile( int nLine, wbArray<int> &aSeq )
		{
			if (m_pFile==NULL) {
				lout_error("file open failed, can't GetLine!!");
			}
			if (nLine >= m_nLine)
				return false;
/*
			long long llpos = (2 + (m_nSize+1) * nLine) * sizeof(int);
			_fseeki64(m_pFile, llpos, SEEK_SET);

			int nLen;
			fread(&nLen, sizeof(int), 1, m_pFile);
			aSeq.SetNum(nLen);
			fread(aSeq.GetBuffer(), sizeof(int), nLen, m_pFile);
*/
			long long llpos = m_llSeqOffset + m_aSeqBegOffset[nLine] * sizeof(int);
			_fseeki64(m_pFile, llpos, SEEK_SET);

			int nLen = m_aSeqBegOffset[nLine+1]-m_aSeqBegOffset[nLine];
			aSeq.SetNum(nLen);
			fread(aSeq.GetBuffer(), sizeof(int), nLen, m_pFile);

			return true;
		}
		/// õnLineеݣ0ʼ
		bool GetLine( int nLine, wbArray<int> &aSeq )
		{
			if ( nLine < aaSeq.GetNum() ) {//ڴжȡ
				aSeq.Copy(*aaSeq[nLine]);
				return true;
			} else {
				return GetLineFromFile( nLine, aSeq );
			}
			return true;
		}
	};

	
}