// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// Chinese corpus processing
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


/**
 * \file 
 * \author wangbin
 * \date 2013-9-23
 * \brief ıزҪΪֽڣUnicodeΪGB2312ʽ
*/
#pragma once

#include "wbSystem.h"

/**
 * \brief ıزҪΪֽڣUnicodeΪGB2312ʽ
*/
namespace wbChTxt
{
	/// жǷΪַ, ĺֺͱ
	inline bool bChChar(const char *str){
		return (*str & 0x80); //GBK:ַĵһֽڵλΪ1
	}
	/// жǷΪı
	inline bool bChMask(const char *str){
		return ( (*str & 0x80) && ((unsigned char)str[0]) <= 9 + 0xA0 ); //GB2312У19λ
	}
	
	const unsigned CharDecode_EnSymNum = 128; ///< ASCIIЧ 128
	const unsigned CharDecode_Space = 10; ///< Ԥ10ֵΪţ<s> </s>.
	inline unsigned Char2Num(const char *info)
	{
		unsigned decoderId = 0;
		if (bChChar(info)) {
			//ӳΪ
			//ڴеı룬Ǳ׼GB2312룬ʹֽС0xA0ֱӵ256*256ıӳΪ
			unsigned h = (unsigned char)info[0];
			unsigned l = (unsigned char)info[1];
			decoderId = h * 256 + l + CharDecode_EnSymNum; //Ӣķ֮
		} else {
			decoderId = *info;
		}
		return decoderId + CharDecode_Space; //ΪʼԤ
	}



	const char c_opt_rep[3] = "=>"; ///< 滻
	const char c_opt_del[3] = "~!"; ///< ɾ

	/// brief ¼ruleļеһ
	struct wbSimpleRule
	{
		wbString field1;  ///< һ
		wbString field2;  ///< ڶ
		wbString opt; ///< 
	};


	/// ַ࣬ԷȽ
	class wbChar
	{
	public:
		char buf[3];
		operator char*() { return buf; }
		bool operator == ( char* p ) { return strcmp(buf, p)==0; }
		bool operator == ( char c ) { return buf[1]=='\0' && buf[0]==c;  }
	};
	/// ַһ仰еַԶжϵַͶַ
	class wbCharIter
	{
	public:
		char *m_pBuf;
		char *m_pCur;
		wbCharIter(char *p) { Init(p); }
		/// ʼ
		void Init(char *p) {
			m_pBuf = p; 
			m_pCur = m_pBuf;
		}
		/// ȡһַŻfalse򷵻true
		bool Next(char *pChar)
		{
			if (*m_pCur=='\0')
				return false;
			
			if (*m_pCur & 0x80) {
				pChar[0] = m_pCur[0];
				pChar[1] = m_pCur[1];
				lout_assert(m_pCur[1] != '\0');
				pChar[2] = '\0';
				m_pCur += 2;
			} else {
				pChar[0] = m_pCur[0];
				pChar[1] = '\0';
				m_pCur += 1;
			}

			return true;
		}
	};


	const char c_units[] = "ʮǧʮǧʮǧ";
	const char c_nums[] = "һ߰˾";
	/// ִתΪִ123 => һٶʮ
	wbString Number2Word(char *pStrNum, int nLen);
	/// תΪ
	wbString Year2Word(char *pStrNum, int nLen);
	/// СתΪ
	wbString Decimals2Word(char *pStrNum, int nLen);
	/// תΪ
	wbString Fraction2Word(char *pStrNum);
	/// ٷתΪ
	wbString Precent2Word(char *pStrNum);
	/// жϾǷǺַ
	bool bSentContentSymbols(char *pStr);


	const char c_ch_year[] = "";
	class wbSentence
	{
	public:
		char *m_pSent;
		wbSentence(char *p):m_pSent(p){};
		bool IsContentZi()
		{
			char c[3];
			wbCharIter iter(m_pSent);
			while (iter.Next(c)) {
				if ( strlen(c) == 2 && !bChMask(c) )
					return true;
			}
			return false;
		}
	};

#define lout_rule lout<<"[rule] "
#define lout_rm lout<<"[RM] "
#define lout_num lout<<"[Num] "

	const char c_en_sentendmark[] = ",.?!:;\"";
	/// ıÿ仰еıԤ
	class TxtFilter
	{
	public:
		wbArray<wbSimpleRule> m_aRules; ///< ַ滻
		wbLHash<const char*, int> m_ruleHash; ///< ҹ

		char *m_pSent; ///< ľ


		wbLHash<const char*, int> m_processedSyms; ///< ѾķŹϣƲҪظӡ
	public:
		TxtFilter():m_pSent(NULL){};

		/// 滻
		void LoadRules(wbFile &file);
		wbString GetSym1( char *&p );

		void SetSent( char *p ) { m_pSent = p; }
		
		/// ʹù
		void PerformRules();

		/// з־
		void SegmentSentence();
		bool bSentEnd(char *c) {
			if ( !bEnSymbol(c) )
				return false;
			for (int i=0; i<strlen(c_en_sentendmark); i++)
				if ( c[0] == c_en_sentendmark[i] )
					return true;
			return false;
		}
		/// Ƴûкֵı
		void RemoveSymbolSegment();
		bool bContentChZi(char *p){
			char c[3];
			wbCharIter iter(p);
			while (iter.Next(c)) {
				if ( strlen(c) == 2 && !bChMask(c) )
					return true;
			}
			return false;
		}
		
		/// ƳעͣСĲ
		void RemovePNote();
		bool bContentSymbol(char *p) {
			char c[3];
			wbCharIter iter(p);
			while (iter.Next(c)) {
				if ( strlen(c) == 1 || bChMask(c) )
					return true;
			}
			return false;
		}
		
		/// Ƴб
		void RemoveMarks();
		bool bEnLetter( char *c ) { return bEnSymbol(c) && ( (c[0]>='a'&&c[0]<='z') || (c[0]>='A'&&c[0]<='Z') ); }

		/// 
		void ProcessNumbers();
		bool bNumber(const char *c) { return c[1]=='\0' && c[0]<='9' && c[0] >='0'; }
		bool bEnSymbol(const char *c) { return c[1]=='\0' && !(c[0] & 0x80); }
		wbString TransNumber( wbString &strNum, const char *nextChar );
	};
}

