// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// corpus class
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


#ifndef _WB_CORPUS_H_
#define _WB_CORPUS_H_

#include "wbLog.h"

static char s_wcharDigit[] = "";
static char s_wcharNumber[] = "һ߰˾";
static char s_wcharUnit[] = "ʮǧʮǧʮǧ";
static char s_wcharPoint[] = "";  //Ԥʹֵ
static char s_wcharColon[] = "ã";

#define WIS_COLON(p) (wbTxtToolkits::WContent(p, s_wcharColon))
#define WIS_POINT(p) (wbTxtToolkits::WContent(p, s_wcharPoint))

class wbTxtToolkits
{
public:
	//Ƚһ˫ֽ
	static bool WEqual(const char *p1, const char *p2) { return (p1[0] == p2[0] && p1[1] == p2[1]) ; }
	static int WEqual(const char *p1, const char *p2, int nlen)
	{
		for (int i=0; 
			i<nlen-1 &&
			*p1 == *p2 &&
			*p1 != '\0' &&
			*p2 != '\0'; 
			i++) 
			{
				++p1;
				++p2;
		}
		return *p1 - *p2;
	}
	//жpWǷpContect
	static bool WContent(const char* pW, const char *pContect) {
		for (int i=0; pContect[i] != '\0'; i+=2) {
			if (WEqual(pContect+i, pW))
				return true;
		}
		return false;
	}
	static void WSet(char *p, const char *pSrc) { p[0] = pSrc[0]; p[1] = pSrc[1]; }
	static int Digit2Int(const char *pStr)
	{
		for (int k=0; k<10; k++)
		{
			if (WEqual(s_wcharDigit+2*k, pStr))
				return k;
		}
		return -1;
	}
	static char *FindDigit(char *pStr, int len=-1) //lenƲҳȣΪҵβΪֹ
	{
		for (int i=0; pStr[i] != '\0' && i != len; i+=2)
		{
			if (Digit2Int(pStr+i) != -1)
				return pStr+i;
		}
		return NULL;
	}

	//ֱתΪ֣1245 = һ
	static void DigitToWord(char *pStr)
	{
		for (int i=0; pStr[i] != '\0'; i+=2)
		{
			int nNum = Digit2Int(pStr+i);
			if (nNum != -1)
			{
				WSet(pStr+i, s_wcharNumber+2*nNum);
			}
		}
	}
	// תΪ ģʽ, 1235 => һǧʮ壻 3.5 => 壬תַڴռҪ㹻
	static void DigitToNum(char *pStr)
	{
		char pWord[1024];

		char *pNumBeg = FindDigit(pStr);
		if (pNumBeg == NULL) //
			return;

		//⴦λ
		if (strlen(pStr) == 2){
			WSet(pStr, s_wcharNumber+2*Digit2Int(pStr));
			return;
		}

		char *pNumEnd = pNumBeg+2;
		for (; *pNumEnd != '\0'; pNumEnd+=2)
		{
			if (Digit2Int(pNumEnd) == -1)
				break;
		}

		int nLen = (pNumEnd-pNumBeg)/2;
		char *pTempCur = pWord;
		for (int i=0; i<nLen; )
		{
			int k = Digit2Int(pNumEnd-2-2*i);
			if (k == 0)
			{
				if ( i == 0) // λ
				{
				}
				else //Ϊ λ
				{
					WSet(pTempCur, s_wcharNumber); //ջ
					pTempCur+=2;
				}

				for(; i<nLen && WEqual(pNumEnd-2-2*i, &s_wcharDigit[0]); ) //ҵΪ0һ
					i++;
			}
			else
			{
				if (i == 0) //λ
				{
					WSet(pTempCur, s_wcharNumber + 2*k); //ջ
					pTempCur += 2;
				}
				else if ( i == 1 && i == nLen-1 && k == 1)
				{
					WSet(pTempCur, s_wcharUnit + 2*i); // ʮ
					pTempCur += 2;
				}
				else
				{
					WSet(pTempCur, s_wcharUnit + 2*i);
					pTempCur += 2;
					WSet(pTempCur, s_wcharNumber + 2*k);
					pTempCur += 2;
				}

				i++;
			}
		}

		//Ϊգ
		if (pTempCur == pWord) {
			WSet(pTempCur, s_wcharNumber);
			pTempCur += 2;
		}

		//pWordеݵ
		char tempWord[3];
		int nWordNum = (pTempCur-pWord)/2;
		for (int i=0; i<nWordNum/2; i++)
		{
			WSet(tempWord, &pWord[2*i]);
			WSet(&pWord[2*i], &pWord[2*(nWordNum-1-i)]);
			WSet(&pWord[2*(nWordNum-1-i)], tempWord);
		}
		//Ӻַ
		strcpy(pTempCur, pNumEnd);

		//жǷΪС
		if (WIS_POINT(pTempCur))
		{
			WSet(pTempCur, "");
			DigitToWord(pTempCur+2);
		}

		//copyԭڴ
		strcpy(pStr, pWord);
	}

	//жǷΪ
	static bool bChSymbol(const char *str )  
	{
		if ( ((unsigned char)str[0]) <= 9 + 0xA0 ) //GB2312У19λ
			return true;
		return false;
	};
	//жǷΪУûз
	static bool bChCharSeq(const char *str)
	{
		int nLen = strlen(str);
		if (nLen%2 == 1)
			return false;

		for (int i=0; i<nLen; i+=2)
		{
			if (str[i] > 0 || bChSymbol(str+i))
				return false;
		}
		return true;
	}


	//ǷתΪȫǷ, Ƿɹ
	static bool TransHalf2Ful(char *pNew, const char *str)
	{
		char *pNewStr = new char[strlen(str)*2+1];
		register char *pCurStr = pNewStr;

		register const char *p = str;
		for (; *p != '\0'; )
		{
			if ((*p) & 0x80) { //ȫַ
				WSet(pCurStr, p);
				p += 2; //һַ
			} else { //ַ
				
				//⴦
				switch(*p)
				{
				case ' ': //ո
					WSet(pCurStr, "");
					break;
				case '.': //
					WSet(pCurStr, "");
					break;
				case '\'': //
					WSet(pCurStr, "");
					break;
				case '"': //˫
					WSet(pCurStr, "");
					break;
				default:
					pCurStr[0] = 163;
					pCurStr[1] = *p + 128;
				}
			
				p += 1; //һַ
			}

			pCurStr += 2;
		}

		*pCurStr = '\0';
		strcpy(pNew, pNewStr);
		delete [] pNewStr;
		return true;
	}

	//ַв
	static char* WFindChar(char *pStr, char c)
	{
		char *p = pStr;
		while(*p != '\0'){
			if ((*p) & 0x80)
				p += 2;
			else {
				if (*p == c)
					return p;
				p += 1;
			}
		}
		return NULL;
	}
};


#endif