// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// File class
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


/**
 * \file 
 * \author wangbin
 * \date 2013-9-23
 * \brief ļ
*/

#ifndef _WB_FILE_H_
#define _WB_FILE_H_

#include "iostream"
#include "fstream"
using namespace std;

#include <stdio.h>
#include <time.h>
#include <cstdarg>	//va_listԶɱprint

#include "wbTitle.h"
#include "wbLog.h"
#include "wbString.h"

//==================  ======================================
#define GS_SENTENCE_LEN	32*1024 ///< 5000,һδַ󳤶
//=====================ļ===================================
#define SAFE_FOPEN(pfile, path, mode) {\
	if ( !path || !(pfile = fopen(path, mode)) ) {cout<<"File Open Failed: "<<path<<endl; exit(0);}} ///< ļ

#define SAFE_FMOPEN(file, path, mode) {\
	file.open(path, mode); if (!file.good()) {cout<<"File Open Failed: "<<path<<endl; exit(0);}} ///< ļ

#define SAFE_FSOPEN(file, path) {\
	file.open(path); if (!file.good()) {cout<<"File Open Failed: "<<path<<endl; exit(0);}} ///< ļ
//=====================ļر===================================
#define SAFE_FCLOSE(fp) { if(fp) {fclose(fp); fp=NULL;} } ///< ļر


//=============== write read ٷ ============================
#define WRITE_VALUE(x,fp) fwrite(&(x),sizeof(x),1,fp)
#define READ_VALUE(x,fp) fread(&(x),sizeof(x),1,fp)
#define FPRINT_VALUE(x, fp) fprintf(fp, #x"=%d\n", (x))
#define FSCANF_VALUE(x, fp) fscanf(fp, #x"=%d\n", &(x))

///ļ
class wbFile
{
public:
	FILE *fp;  ///< ļָ
	int nLine; ///< 1ʼ

	char *pStrLine; ///< ַbuffer
	wbString strFileName; ///< ļƣҪΪܹReopenļ

public:
	wbFile():fp(NULL),pStrLine(NULL) { fp = NULL; }
	wbFile(const char *path, const char *mode, bool bHardOpen = true):fp(NULL),pStrLine(NULL) { Open(path, mode, bHardOpen); }
	~wbFile() { SAFE_FCLOSE(fp); delete []pStrLine; }

	virtual bool Open(const char *path, const char *mode, bool bHardOpen = true) 
	{ 
		Close();
		if (path == NULL || mode == NULL)
			return false;

		if ( bHardOpen ) {
			SAFE_FOPEN(fp, path, mode);
		} else {
			fp = fopen(path, mode);
		}
		
		strFileName = path;
		nLine = 0; 
		return(fp!=NULL); 
	}
	/// ´ļ
	virtual bool Reopen( const char* model )
	{
		return Open( strFileName, model );
	}
	virtual void Close() { SAFE_FCLOSE(fp); }

	operator FILE* () { return fp; }

	/// һУڵbufferַָ룬
	/**
	 * \param [in] bPrecent ΪtruecmdtitleӡļȣĬfalse
	 * \return ַָ룬ֱchar*գbufferά
	*/
	virtual char *GetLine(bool bPrecent = false)
	{
		if (!Good()) {
			lout_error("file open failed, can't GetLine!!");
		}

		if (bPrecent && nLine == 0)
			wbTitle::Precent(fp, true, "Read File ");

		if (!pStrLine)
			pStrLine = new char[GS_SENTENCE_LEN];

		bool bOk = GetLine(pStrLine, GS_SENTENCE_LEN);
		if (bPrecent)
			wbTitle::Precent(fp);


		if ( bOk )
			return pStrLine;
		else
			return NULL;
	}
	
	/// һеstrУbufferⲿ٣ʹ
	virtual bool GetLine(char *str, int maxLen = GS_SENTENCE_LEN) 
	{
		char *p = fgets(str, maxLen, fp);
		if (p == NULL)
			return false;

		nLine++; //βnLine
		int l = strlen(str);
		if (l >= maxLen-1)
		{
			cout<<"[FILE GetLine], (nline="<<nLine<<") ַԤٿռ"<<endl;
			return false;
		}
		if (str[l-1] == '\n')
			str[l-1] = '\0';

		return true; 
	}
	
	///ӡʽӡ
	virtual void Print(const char* p_pMessage,...)
	{
		if (fp == NULL)
			return;

		char strBuffer[GS_SENTENCE_LEN];
		va_list vaParams;
		va_start(vaParams,p_pMessage);

		_vsnprintf(strBuffer,GS_SENTENCE_LEN,p_pMessage,vaParams);

		fprintf(fp, "%s", strBuffer);

		fflush(fp);
	}

	/// ջ棬fflush
	void Flush()  { fflush(fp); }
	/// λ
	/** ʹ_fseeki64λãֹļʹfseekʧЧ */
	void Reset() { _fseeki64(fp,0, SEEK_SET); nLine=0; }
	/// ļǷ
	bool Good() { return !(fp==NULL);}
	/// ӡ
	template <typename TYPE>
	void PrintArray( char* pformat, TYPE* pbuf, int num )
	{
		for (int i=0; i<num; i++) {
			Print(pformat, *pbuf);
			putc(' ', fp);
			pbuf++;
		}
		putc('\n', fp);
	}
};

/// wbComFile,ͬʱļĻӡݣwbLog󼸺ʹ
class wbComFile : public wbFile
{
private:
	bool m_bOutputCmd;
	bool m_bOutputFile;
public:
	wbComFile():wbFile() { m_bOutputCmd = true; m_bOutputFile = true; };
	wbComFile(const char *path, const char *mode) : wbFile(path, mode) {m_bOutputCmd = true; m_bOutputFile = true;};
	void OutputCmd(bool bOut) { m_bOutputCmd = bOut; }
	void OutputFile(bool bOut) { m_bOutputFile = bOut; }
	virtual void Print(const char* p_pMessage,...)
	{
		char strBuffer[GS_SENTENCE_LEN];
		va_list vaParams;
		va_start(vaParams,p_pMessage);

		_vsnprintf(strBuffer,GS_SENTENCE_LEN,p_pMessage,vaParams);

		if (fp && m_bOutputFile)
			fprintf(fp, "%s", strBuffer);

		if (m_bOutputCmd)
			printf("%s", strBuffer);
	}
};

/// obj࣬ṩ˺дĽӿڣ
class wbObj
{
public:
	virtual void WriteT(wbFile &file) {};
	virtual void ReadT(wbFile &file) {};
	virtual void WriteB(wbFile &file) {};
	virtual void ReadB(wbFile &file) {};
};

/// wbObjFileдͬObjʱʹ
class wbObjFile : public wbFile
{
public:
	int m_nTotalNum; ///< Objܸ
	int m_nCurNum;   ///< ǰObj

	wbObj *m_pObj; ///< Objָ룬ÿָָObjж

	wbObjFile():wbFile(),m_nTotalNum(0),m_nCurNum(-1){}
	wbObjFile(const char *path, const char *mode):wbFile(path, mode),m_nTotalNum(0),m_nCurNum(-1){}

	/// txtдļͷ
	void WriteHeadT()
	{
		Print("nTotalNum = %d\n\n", m_nTotalNum);
		m_nCurNum = 0;
	}
	/// txtļͷ
	void ReadHeadT()
	{
		fscanf(wbFile::fp, "nTotalNum = %d\n\n", &m_nTotalNum);
		m_nCurNum = 0;
	}

	/// txtдObj
	void WriteObjT()
	{
		Print("nObj: %d\n", m_nCurNum);
		m_pObj->WriteT(*this);
		Print("\n");
		m_nCurNum++;
	}
	/// txtObj
	bool ReadObjT()
	{
		if (feof(fp))
			return false;
		if (m_nCurNum >= m_nTotalNum)
			return false;

		fscanf(wbFile::fp, "nObj: %d\n", &m_nCurNum);

		m_pObj->ReadT(*this);
		fscanf(wbFile::fp, "\n");
		return true;
	}

	/// binдļͷ
	void WriteHeadB()
	{
		fwrite(&m_nTotalNum, sizeof(m_nTotalNum), 1, fp);
		m_nCurNum = 0;
	}
	/// binļͷ
	void ReadHeadB()
	{
		fread(&m_nTotalNum, sizeof(m_nTotalNum), 1, fp);
		m_nCurNum = 0;
	}
	/// binдObj
	void WriteObjB()
	{
		fwrite(&m_nCurNum, sizeof(m_nCurNum), 1, fp);
		m_pObj->WriteB(*this);
		m_nCurNum++;
	}
	/// binObj
	bool ReadObjB()
	{
		if (feof(fp))
			return false;
		if (m_nCurNum >= m_nTotalNum)
			return false;

		fread(&m_nCurNum, sizeof(m_nCurNum), 1, fp);
		m_pObj->ReadB(*this);
		return true;
	}
};




#endif