// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// Gradient descent algorithms
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


#pragma once
#include "wbLBFGS.h"

#define lout_GD lout<<"[LBFGS] "
#define lout_GD_iter if (m_nIterNum%m_opt_nPrintPerIter==0) lout_LBFGS
#define lout_GD_print {\
	lout_GD<<"a="<<dStep<<" f_k="<<dPreValue<<" f_(k+1)="<<dCurValue<<endl; \
	lout_GD<<"df_norm="<<dCurGradNorm<<" dx_norm="<<dParamDist<<" f="<<dValueDist<<endl; \
	lout_GD<<"ExValues={ "; \
	for (int mm=0; mm<nExValueNum; mm++) lout<<dExValues[mm]<<" "; \
	lout<<"}"<<endl; \
	 }


/** 
 * \class wbGD
 * \author wangbin
 * \date 2014-9-23
 * \brief 
 */
class wbGD
{
public:
	wbFunc *m_pFunc; ///< Ŀ꺯
	
	int m_nIterNum; ///< ǰ
	double m_dStep; ///< 
	double m_dTolerance; ///< ݲֹݶСݲʱ
	double *m_pdResParams; ///< 

	double m_dSpendMinute; ///< ¼ѵʱ䣨ӣ

	int m_opt_nMaxIterNum; ///< Ϊ0ֱ
	int m_opt_nPrintPerIter; ///< ÿεӡһνĬ1Ϊ0򲻴ӡ
	bool m_opt_bFixStepLineSearch; ///< Ƿ̶

	wbFile *m_pfdbg_dir; ///< ӡûx
	wbFile *m_pfdbg_x; ///< ӡÿεxֵ

public:
	wbGD(wbFunc *pfunc = NULL):m_pFunc(pfunc)
	{
		m_nIterNum = 0;
		m_dStep = 0.1;
		m_dTolerance = 1e-5;
		m_pdResParams = NULL;
		
		m_dSpendMinute = 0;
		m_opt_nMaxIterNum = -1;
		m_opt_nPrintPerIter = 1;

		m_pfdbg_dir = NULL;
		m_pfdbg_x = NULL;
		m_opt_bFixStepLineSearch = false;
	}

	~wbGD()
	{
		SAFE_DELETE_ARRAY(m_pdResParams);
		SAFE_DELETE(m_pfdbg_dir);
		SAFE_DELETE(m_pfdbg_x);
	}

	/// ݶ½㷨
	/** 
	 * \param [in] pdInitParam =NULL趨ֵ
	 * \return ɹ򷵻true򷵻false
	 */
	bool Run(double *pdInitParam = NULL)
	{
		wbClock ck;

		if (!m_pFunc) {
			lout_LBFGS<<"m_pFunc == NULL"<<endl;
			return false;
		}
		SAFE_DELETE_ARRAY(m_pdResParams);
		m_pdResParams = new double[m_pFunc->m_nParamNum];

		double *pdCurParams = new double[m_pFunc->m_nParamNum]; //ǰx_k
		double *pdPreParams = new double[m_pFunc->m_nParamNum]; //ǰһεĲ x_(k-1)
		double *pdCurGradient = new double[m_pFunc->m_nParamNum]; //ǰݶ df_k
		double *pdPreGradient = new double[m_pFunc->m_nParamNum]; //ǰһݶ df_(k-1)
		double dCurValue = 0; // ֵ f_k
		double dPreValue = 0; // ֵ f_(k-1)
		double *pdDir = new double[m_pFunc->m_nParamNum]; //,p_k
		double nExValueNum; // ݵĴС
		double dExValues[wbFunc_nExValueMaxNum]; // 

		double dStep = 0; ///< 
		double dCurGradNorm = 0; //ǰݶȵģ
		double dPreGradNorm = 0; //һݶȵģ
		double dParamDist = 0; //εĲ
		double dValueDist = 0; //ֵ
		m_dSpendMinute = 0; //ʱΪ0

		//ʼ
		for (int i=0; i<m_pFunc->m_nParamNum; i++) {
			pdCurParams[i] = (pdInitParam)? pdInitParam[i]: 1;
			pdPreParams[i] = 0;
			pdCurGradient[i] = 0;  //ݶȳʼ
			pdPreGradient[i] = 0;  //ݶȳʼ
		}
		//ʼݶȺͺֵ
		dCurValue = m_pFunc->GetValue(pdCurParams);
		nExValueNum = m_pFunc->GetExtraValues(pdCurParams, dExValues);
		lout_GD_iter<<"========== iter:Init === ("<<m_dSpendMinute<<"m) ======="<<endl;
		lout_GD_print;
		m_pFunc->GetGradient(pdCurParams, pdCurGradient);
		dCurGradNorm = VecNorm(pdCurGradient, m_pFunc->m_nParamNum);

		//ʼ
		for (m_nIterNum=0; m_opt_nMaxIterNum !=0 ;m_nIterNum++)
		{
			ck.Begin();

			lout_GD_iter<<"======== iter:"<<m_nIterNum<<" ===("<<m_dSpendMinute<<"m)======="<<endl;

			//p_k
			ComputeDir(m_nIterNum, pdCurGradient, pdPreGradient, dCurGradNorm, dPreGradNorm, pdDir);
			lout_GD_iter<<"df_k = {"<<pdCurGradient[0]<<","<<pdCurGradient[1]<<"...}"<<endl;
			lout_GD_iter<<"p_k  = {"<<pdDir[0]<<","<<pdDir[1]<<"...}"<<endl;

			//һά
			Repeat(pdCurParams, pdPreParams);
			Repeat(pdCurGradient, pdPreGradient);
			dPreValue = dCurValue;
			dPreGradNorm = dCurGradNorm;
			double dStep = LineSearch(pdDir, dPreValue, pdPreGradient, pdPreParams, pdCurParams);

			//µݶȺͺֵ
			lout_LBFGS_iter<<"x_k+1 = {"<<pdCurParams[0]<<","<<pdCurParams[1]<<"...}"<<endl;
			dCurValue = m_pFunc->GetValue(pdCurParams);
			m_pFunc->GetGradient(pdCurParams, pdCurGradient);


			dCurGradNorm = VecNorm(pdCurGradient, m_pFunc->m_nParamNum);//ݶȵģ
			dParamDist = VecDist(pdPreParams, pdCurParams, m_pFunc->m_nParamNum); //εĲ
			dValueDist = fabs(dPreValue - dCurValue); //ֵ

			//ӡϢ
			if ((m_nIterNum+1)%m_opt_nPrintPerIter==0) {
				nExValueNum = m_pFunc->GetExtraValues(pdCurParams, dExValues); //ӡϢǲŻ
				lout_GD_print;
			}
			//ӡļ
			OutputDbg(pdCurParams, pdCurGradient, pdDir, dStep);

			//ֹж
			if ( dCurGradNorm <= m_dTolerance ) 
			{
				lout_GD<<"========= [Succeed] iter:"<<m_nIterNum<<" ========="<<endl;
				lout_GD_print;
				break;
			}

			if ( m_opt_nMaxIterNum>0 && (m_nIterNum+1) >= m_opt_nMaxIterNum ) 
			{
				lout_GD<<"========= [MaxIterNum] iter:"<<m_nIterNum<<" ========="<<endl;
				lout_GD_print;
				break;
			}

			m_dSpendMinute += ck.ToSecond( ck.End() ) / 60;

			//Pause();
		}

		//
		memcpy( m_pdResParams, pdCurParams, sizeof(m_pdResParams[0])*(m_pFunc->m_nParamNum) );

		return true;
	}

	///㷽p_k
	/**
	 * \param [in] k 
	 * \param [in] pdCurGradient x_kݶֵ
	 * \param [in] pdPreGradient x_k-1ݶֵ
	 * \param [out] pdDir Żؼķ
	*/
	virtual void ComputeDir(int k, double *pdCurGradient, double *pdPreGradient, double &dCurGradNorm, double &dPreGradNorm, double *pdDir)
	{
		for (int i=0; i<m_pFunc->m_nParamNum; i++)
			pdDir[i] = - pdCurGradient[i];
	}

	virtual void OutputDbg(double *pParams, double *pGradient, double *pdDir, double dStep)
	{
		if ( m_pfdbg_dir ) {
			for (int i=0; i<m_pFunc->m_nParamNum; i++)
				m_pfdbg_dir->Print("%f ", dStep * pdDir[i]);
			m_pfdbg_dir->Print("\n");
		}
		if ( m_pfdbg_x ) {
			for (int i=0; i<m_pFunc->m_nParamNum; i++)
				m_pfdbg_x->Print("%f ", pParams[i]);
			m_pfdbg_x->Print("\n");
		}
	}

	/// һά, ˴򵥵
	/** 
	 * \param [in] pdDir p_k
	 * \param [in] dValue x_kĺֵ
	 * \param [in] pdGradient x_kݶ
	 * \param [in] pdCurParam x_k
	 * \param [out] pdNextParam ͬʱx_(k+1)
	 * \return  ز
	 */
	virtual double LineSearch(double *pdDir, double dValue, double *pdGradient, double *pdCurParam, double *pdNextParam)
	{
		if ( m_opt_bFixStepLineSearch ) 
		{
			for (int i=0; i<m_pFunc->m_nParamNum; i++)
				pdNextParam[i] = pdCurParam[i] + m_dStep * pdDir[i];
			return 1;
		}

		/// һά
		/// 㷨μNumerical OptimizationP57ܵinterpolation㷨
		double a0=0, a1=0, a2=0; //
		double phi0=0, phi1=0, phi2=0; //phi(ai)
		double c = 1e-4;
		double phi_t = VecProduct(pdGradient, pdDir, m_pFunc->m_nParamNum); // phi'(0)

		a2 = m_dStep; //ʼΪ1
		for (int k=1; a2>0 ;k++)
		{
			// x = x + a * p
			for (int n=0; n<m_pFunc->m_nParamNum; n++)
				pdNextParam[n] = pdCurParam[n] + a2 * pdDir[n];

			phi2 = m_pFunc->GetValue(pdNextParam); // phi(a2)

			if (phi2 <= dValue + c * a2 * phi_t)
				break;

			//ǰεĽ
			a0 = a1;
			a1 = a2;
			phi0 = phi1;
			phi1 = phi2;
			if ( k== 1 ) {
				a2 = - phi_t*a1*a1 / 2 / ( phi1 - dValue - phi_t * a1 );
			} else {
				double v1 = phi1 - dValue - phi_t * a1;
				double v2 = phi0 - dValue - phi_t * a0;
				double a = a0*a0*v1 - a1*a1*v2;
				double b = -a0*a0*a0*v1 + a1*a1*a1*v2;
				double t = a0*a0*a1*a1*(a1-a0);
				a /= t;
				b /= t;
				a2 = (-b+sqrt(b*b-3*a*phi_t)) / (3*a);
			}

			//a2a1̫ӽ̫Զȡa2=a1/2
			if (fabs(a2-a1) < 1e-5 ||
				a1/a2 > 10 )
				a2 = a1/2;
			//a2Сȡһ̫Сֵ
			if (a2 < 1e-5) {
				lout_warning("a2 is too small < 1e-5, break")
					a2 = 1e-5;
				break;
			}
		}
		return a2;
		
	}

	/// ڻ
	static double VecProduct(double *pdVec1, double *pdVec2, int nSize)
	{
		double d = 0;
		for (int i=0; i<nSize; i++) 
			d += pdVec1[i] * pdVec2[i];
		return d;
	}
	/// ģ
	static double VecNorm(double *pdVec, int nSize)
	{
		return sqrt(VecProduct(pdVec, pdVec, nSize));
	}
	/// ľ
	static double VecDist(double *pdVec1, double *pdVec2, int nSize)
	{
		double d = 0;
		for (int i=0; i<nSize; i++)
			d += pow(pdVec1[i]-pdVec2[i], 2);
		return sqrt(d);
	}
	/// жǷΪ0
	static bool VecZero(double *pdVec, int nSize)
	{
		for (int i=0; i<nSize; i++) {
			if (fabs(pdVec[i]) > 1e-10) 
				return false;
		}
		return true;
	}

	/// ָ
	void Repeat(double *&p1, double *&p2)
	{
		double *pTemp = p1;
		p1 = p2;
		p2 = pTemp;
	}
};



/** 
 * \class wbGDOpt
 * \author wangbin
 * \date 2014-9-23
 * \brief Zhiqiang㷨ĸĽݶ½
 */
 class wbGDOpt : public wbGD
 {
 public:
	 double m_tol; ///< ֵжݶǷ
	 wbGDOpt(wbFunc *pfunc = NULL) : wbGD(pfunc)
	 {
		 m_tol = 1e-3;
	 }

	 virtual void ComputeDir(int k, double *pdCurGradient, double *pdPreGradient, double &dCurGradNorm, double &dPreGradNorm, double *pdDir)
	 {

		 for (int i=0; i<m_pFunc->m_nParamNum; i++)
			 pdDir[i] = pdCurGradient[i];

		 if ( m_nIterNum > 0 ) {  //һ
			 double sq1 = dPreGradNorm * dPreGradNorm;
			 double sq2 = dCurGradNorm * dCurGradNorm;
			 double cr = VecProduct(pdCurGradient, pdPreGradient, m_pFunc->m_nParamNum);

			 double r = 1 - cr * cr / (sq1 * sq2);
			 if ( r < m_tol ) 
			 {
				 double lam = cr / sq1 - 1;
				 if (lam < 0) {
					 for (int i=0; i<m_pFunc->m_nParamNum; i++)
						 pdDir[i] = - pdDir[i] / lam;
				 }
			 }
		 }

		 for (int i=0; i<m_pFunc->m_nParamNum; i++)  //ȡ
			 pdDir[i] = -pdDir[i];
	 }
 };
