// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// LBFGS algorithms
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.



/**
 * \file
 * \author wangbin
 * \date 2013-11-13
 * \brief Quasi-Newton MethodLBFGS method
 */
#pragma once
#include "wbSystem.h"

const int wbFunc_nExValueMaxNum = 100;///< صĶ
/** 
 * \class wbFunc
 * \author wangbin
 * \date 2013-11-13
 * \brief function࣬ŻĿ꺯Ҫ
 */
class wbFunc
{
public:
	int m_nParamNum; ///< ʺм
public:
	wbFunc(int nParamNum = 0):m_nParamNum(nParamNum){};
	/// Ŀ꺯pdParamsȡֵ
	virtual double GetValue(double *pdParams) = 0;
	/// Ŀ꺯pdParamsݶpdGradientСҪͲͬ
	virtual bool GetGradient(double *pdParams, double *pdGradient) = 0;
	/// ӿڵзŻһֵԼȻڵдӡļ
	/** pdValuesڴⲿ٣ݸwbFunc_nExValueMaxNum */
	virtual int GetExtraValues(double *pdParams, double *pdValues) { return 0; }
};



#define lout_LBFGS lout<<"[LBFGS] "
#define lout_LBFGS_iter if (m_nIterNum%m_opt_nPrintPerIter==0) lout_LBFGS
#define lout_LBFGS_print {\
							lout_LBFGS<<"a="<<dStep<<" "<<"f_k="<<dPreValue<<" f_(k+1)="<<dCurValue<<endl; \
							lout_LBFGS<<"df_norm="<<dGradNorm<<" dx_norm="<<dParamDist<<" f="<<dValueDist<<endl; \
							lout_LBFGS<<"ExValues={ "; \
							for (int mm=0; mm<nExValueNum; mm++) lout<<dExValues[mm]<<" "; \
							lout<<"}"<<endl; \
						 }

typedef enum { 
	grad, ///< ݶȣжݶȵģǷӽ0
	func, ///< ֵжκֵ֮Ƿ㹻С
	param  ///< Ƿ㹻С
} wbOptTerminal; ///< ͣŻֹ

/** 
 * \class wbLBFGS
 * \author wangbin
 * \date 2013-11-13
 * \brief LBFGS㷨,ҪṩwbFuncָΪĿ꺯
 */
class wbLBFGS
{
	typedef struct {
		double *s;
		double *y;
	} sy; 
public:
	wbFunc *m_pFunc; ///< Ŀ꺯

	int m_nLimitiedNum; ///< ʷm
	double m_dTolerance; ///< ݲֹݶСݲʱ
	wbOptTerminal m_terminal; ///< ֹжϵ

	int m_nIterNum; ///< ǰ
	
	sy *m_pCirQueueBuf; ///< ѭbuffers_k = x_k - x_(k-1)  y_k ݶ
	int m_nCirQueueBufTail; ///< β
	double *m_pd_s, *m_pd_y; ///< 浱ǰº s_k = x_k - x_{k-1} and y_k = ݶȲ

	double *m_pdResParams; ///< 

	double m_dSpendMinute; ///< ¼ѵʱ䣨ӣ
	//[opt]
	int m_opt_nIterBeg; ///< Ӵֵʼѭ
	double m_opt_dFixedStep; ///< ̶Ϊ0(һ0~1)̶GetValueķֵãӰ
	bool m_opt_bGradientDescent; ///< ֱʹʹøݶΪ½
	int m_opt_nMaxIterNum; ///< <0ֱ
	int m_opt_nLineSearchMaxIterNum; ///< 
	int m_opt_nPrintPerIter; ///< ÿεӡһνĬ1Ϊ0򲻴ӡ
	FILE* m_opt_pOutputIter; ///< ΪNULLļдӡϢ
	int m_opt_nSetRand; ///< ڵпܻõڲÿm_opt_nSetRand,ĬΪ1000
	bool m_opt_bEndIter; ///< Ƶֹ
	wbFile *m_pfdbg_dir; ///< ӡûx
	wbFile *m_pfdbg_x; ///< ӡÿεX
	wbFile *m_pfdbg_g; ///< ӡÿεݶ

	wbLBFGS(wbFunc *pfunc = NULL):m_pFunc(pfunc)
	{
		m_nLimitiedNum = 7;
		m_dTolerance = 1e-5;
		m_terminal = grad;
		m_nIterNum = 0;

		m_pCirQueueBuf = NULL;
		m_nCirQueueBufTail = 0;
		m_pd_s = NULL;
		m_pd_y = NULL;

		m_pdResParams = NULL;

		m_dSpendMinute = 0;

		m_opt_nIterBeg = 0;
		m_opt_dFixedStep = 0;
		m_opt_bGradientDescent = false;
		m_opt_nMaxIterNum = -1;
		m_opt_nLineSearchMaxIterNum = 0;
		m_opt_nPrintPerIter = 1;
		m_opt_nPrintPerIter = NULL;
		m_opt_pOutputIter = NULL;
		m_opt_nSetRand = 1000;
		m_opt_bEndIter = false;
		m_pfdbg_dir = NULL;
		m_pfdbg_x = NULL;
		m_pfdbg_g = NULL;
	}
	~wbLBFGS()
	{
		SAFE_DELETE_ARRAY(m_pdResParams);

		SAFE_DELETE(m_pfdbg_dir);
		SAFE_DELETE(m_pfdbg_x);
		SAFE_DELETE(m_pfdbg_g);
	}

	/// LBFGS㷨
	/** 
	 * \param [in] pdInitParam =NULL趨ֵ
	 * \return ɹ򷵻true򷵻false
	 */
	bool Run(double *pdInitParam = NULL)
	{
		wbClock ck;

		if (!m_pFunc) {
			lout_LBFGS<<"m_pFunc == NULL"<<endl;
			return false;
		}
		SAFE_DELETE_ARRAY(m_pdResParams);
		m_pdResParams = new double[m_pFunc->m_nParamNum];
		
		double *pdCurParams = new double[m_pFunc->m_nParamNum]; //ǰx_k
		double *pdPreParams = new double[m_pFunc->m_nParamNum]; //ǰһεĲ x_(k-1)
		double *pdCurGradient = new double[m_pFunc->m_nParamNum]; //ǰݶ df_k
		double *pdPreGradient = new double[m_pFunc->m_nParamNum]; //ǰһݶ df_(k-1)
		double dCurValue = 0; // ֵ f_k
		double dPreValue = 0; // ֵ f_(k-1)
		double *pdDir = new double[m_pFunc->m_nParamNum]; //,p_k
		double *pdAlpha = new double[m_nLimitiedNum + 1]; //p_kʱʱalpha
		double nExValueNum; // ݵĴС
		double dExValues[wbFunc_nExValueMaxNum]; // 

		double dStep = 0; ///< 
		double dGradNorm = 0; //ݶȵģ
		double dParamDist = 0; //εĲ
		double dValueDist = 0; //ֵ

		m_dSpendMinute = 0; //ʱΪ0

		//ʼ
		for (int i=0; i<m_pFunc->m_nParamNum; i++) {
			pdCurParams[i] = (pdInitParam)? pdInitParam[i]: 1;
			pdPreParams[i] = 0;
		}
		//ʼݶȺͺֵ
		dCurValue = m_pFunc->GetValue(pdCurParams);
		nExValueNum = m_pFunc->GetExtraValues(pdCurParams, dExValues);
		lout_LBFGS_iter<<"========== iter:Init === ("<<m_dSpendMinute<<"m) ======="<<endl;
		lout_LBFGS_print;
		m_pFunc->GetGradient(pdCurParams, pdCurGradient);
		//GetGradient(pdCurParams, pdCurGradient);


		//ʼѭ
		CirQueueBuf_Init();

		//ĳʼ
		IterInit();
		
		//ʼ
		for (m_nIterNum=m_opt_nIterBeg; m_opt_nMaxIterNum !=0 ;m_nIterNum++)
		{
			ck.Begin();

			lout_LBFGS_iter<<"======== iter:"<<m_nIterNum<<" ===("<<m_dSpendMinute<<"m)======="<<endl;
			//srand
			if (m_nIterNum%m_opt_nSetRand==0)
				srand(time(NULL));

			//p_k
			ComputeDir(m_nIterNum, pdCurGradient, pdAlpha, pdDir);
			lout_LBFGS_iter<<"df_k = {"<<pdCurGradient[0]<<","<<pdCurGradient[1]<<"...}"<<endl;
			lout_LBFGS_iter<<"p_k  = {"<<pdDir[0]<<","<<pdDir[1]<<"...}"<<endl;

			//һά
			Repeat(pdCurParams, pdPreParams);
			Repeat(pdCurGradient, pdPreGradient);
			dPreValue = dCurValue;
			dStep = LineSearch(pdDir, dPreValue, pdPreGradient, pdPreParams, pdCurParams);

			//µݶȺͺֵ
			lout_LBFGS_iter<<"x_k+1 = {"<<pdCurParams[0]<<","<<pdCurParams[1]<<"...}"<<endl;
			dCurValue = m_pFunc->GetValue(pdCurParams);
			//m_pFunc->GetGradient(pdCurParams, pdCurGradient);
			GetGradient(pdCurParams, pdCurGradient);
			
			
			dGradNorm = VecNorm(pdCurGradient, m_pFunc->m_nParamNum);//ݶȵģ
			dParamDist = VecDist(pdPreParams, pdCurParams, m_pFunc->m_nParamNum); //εĲ
			dValueDist = fabs(dPreValue - dCurValue); //ֵ

			//ӡϢ
			if ((m_nIterNum+1)%m_opt_nPrintPerIter==0) 
			{
				nExValueNum = GetExValues(pdCurParams, dExValues); //ӡϢǲŻ
				lout_LBFGS_print;
			
				if (m_opt_pOutputIter) { /// ӡļ
					fprintf(m_opt_pOutputIter, "%d\t%f %f ", m_nIterNum, dCurValue, dGradNorm);
					for (int mm=0; mm<nExValueNum; mm++)
						fprintf(m_opt_pOutputIter, "%f ", dExValues[mm]);
					fprintf(m_opt_pOutputIter, "\n");
					fflush(m_opt_pOutputIter);
				}
				OutputDbg(pdCurParams, pdCurGradient, pdDir);
			}

			//ֹж
			if ( (m_terminal==grad && dGradNorm <= m_dTolerance ) ||
				( m_terminal==func && dValueDist <= m_dTolerance) || 
				( m_terminal==param && dParamDist <= m_dTolerance) ||
				m_opt_bEndIter ) 
			{
				lout_LBFGS<<"========= [Succeed] iter:"<<m_nIterNum<<" ========="<<endl;
				lout_LBFGS_print;
				break;
			}

			if ( m_opt_nMaxIterNum>0 && (m_nIterNum+1) >= m_opt_nMaxIterNum ) 
			{
				lout_LBFGS<<"========= [MaxIterNum] iter:"<<m_nIterNum<<" ========="<<endl;
				lout_LBFGS_print;
				break;
			}


			//浽
			//double *pd_s, *pd_y;
//			CirQueueBuf_In(m_pd_s, m_pd_y);
			AddDiff(pdCurParams, pdPreParams, pdCurGradient, pdPreGradient);
// 			for (int n=0; n<m_pFunc->m_nParamNum; n++) {
// 				m_pd_s[n] = pdCurParams[n] - pdPreParams[n];
// 				m_pd_y[n] = pdCurGradient[n] - pdPreGradient[n];
// 			}

			IterStepEnd(m_nIterNum);

			m_dSpendMinute += ck.ToSecond( ck.End() ) / 60;
		}

		//
		memcpy( m_pdResParams, pdCurParams, sizeof(m_pdResParams[0])*(m_pFunc->m_nParamNum) );

		CirQueueBuf_Release();
		SAFE_DELETE_ARRAY(pdCurParams);
		SAFE_DELETE_ARRAY(pdPreParams);
		SAFE_DELETE_ARRAY(pdCurGradient);
		SAFE_DELETE_ARRAY(pdPreGradient);
		SAFE_DELETE_ARRAY(pdDir);
		SAFE_DELETE_ARRAY(pdAlpha);

		IterEnd();

		return true;
	}

	///< ʼ
	virtual void IterInit() {}

	virtual void OutputDbg(double *pParams, double *pGradient, double *pdDir)
	{
		if ( m_pfdbg_dir ) {
			for (int i=0; i<m_pFunc->m_nParamNum; i++)
				m_pfdbg_dir->Print("%f ", pdDir[i]); 
			m_pfdbg_dir->Print("\n");
		}
		if ( m_pfdbg_x ) {
			for (int i=0; i<m_pFunc->m_nParamNum; i++)
				m_pfdbg_x->Print("%f ", pParams[i]);
			m_pfdbg_x->Print("\n");
		}
		if ( m_pfdbg_g ) {
			for (int i=0; i<m_pFunc->m_nParamNum; i++)
				m_pfdbg_g->Print("%f ", pGradient[i]);
			m_pfdbg_g->Print("\n");
		}
	}

	///< öҪӡĲ
	virtual int GetExValues( double *pdCurParams, double *pdExValue )
	{
		return m_pFunc->GetExtraValues( pdCurParams, pdExValue );
	}

	///< Gradient
	virtual void GetGradient(double *pdCurParam, double *pdCurGradient, double *pdPreGradient = NULL)
	{
		m_pFunc->GetGradient(pdCurParam, pdCurGradient);
	}

	///< sy
	virtual void AddDiff(double *pdCurParams, double *pdPreParams, double *pdCurGradient, double *pdPreGradient)
	{
		CirQueueBuf_In(m_pd_s, m_pd_y);
		for (int n=0; n<m_pFunc->m_nParamNum; n++) {
			m_pd_s[n] = pdCurParams[n] - pdPreParams[n];
			m_pd_y[n] = pdCurGradient[n] - pdPreGradient[n];
		}
	}

	/// һЩ⴦
	virtual void IterStepEnd(int k) {};

	/// ȫЩ⴦
	virtual void IterEnd() {};

	/// ѭгʼ
	void CirQueueBuf_Init()
	{
		m_pCirQueueBuf = new sy[m_nLimitiedNum];
		for (int i=0; i<m_nLimitiedNum; i++) {
			m_pCirQueueBuf[i].s = new double[m_pFunc->m_nParamNum];
			m_pCirQueueBuf[i].y = new double[m_pFunc->m_nParamNum];
		}
		m_nCirQueueBufTail = 0;
	}
	/// ѭͷ
	void CirQueueBuf_Release()
	{
		if ( m_pCirQueueBuf ) {
			for (int i=0; i<m_nLimitiedNum; i++) {
				SAFE_DELETE_ARRAY(m_pCirQueueBuf[i].s);
				SAFE_DELETE_ARRAY(m_pCirQueueBuf[i].y);
			}
			SAFE_DELETE_ARRAY(m_pCirQueueBuf);
		}
		m_nCirQueueBufTail = 0;
	}
	
	/// ѭеǰi,i<=m_nLimitiedNum
	void CirQueueBuf_Prev(int i, double *&pd_s, double *&pd_y)
	{
		i = (m_nLimitiedNum + m_nCirQueueBufTail - i) % m_nLimitiedNum;
		pd_s = m_pCirQueueBuf[i].s;
		pd_y = m_pCirQueueBuf[i].y;
	}

	/// ӣtailƣͬʱضβ
	void CirQueueBuf_In(double *&pd_s, double *&pd_y)
	{
		pd_s = m_pCirQueueBuf[m_nCirQueueBufTail].s;
		pd_y = m_pCirQueueBuf[m_nCirQueueBufTail].y;
		m_nCirQueueBufTail = (m_nCirQueueBufTail+1) % m_nLimitiedNum;
	}

	///㷽p_k㷨οNumerical OptimizationP178Algorithm 7.4
	/**
	 * \param [in] k 
	 * \param [in] pdGradient x_kݶֵ
     * \param [in] pdAlpha мڴռ䣬СΪm_nLimitiedNumÿζҪ٣˲ⲿķʽ
	 * \param [out] pdDir Żؼķ
	*/
	virtual void ComputeDir(int k, double *pdGradient, double *pdAlpha, double *pdDir)
	{
		double *pd_s = NULL;
		double *pd_y = NULL;
		int nVecLen = m_pFunc->m_nParamNum;
		

		if (m_opt_bGradientDescent) {  //ʹݶΪ½
			//Ҫȡ
			for (int n=0; n<nVecLen; n++)
				pdDir[n] = -pdGradient[n];
			return ;
		}

		lout_LBFGS<<"LBFGS dir computer"<<endl;
		int nBound = min(m_nLimitiedNum, k); //ֻǰm

		//ݶȸq
		memcpy(pdDir, pdGradient, sizeof(pdDir[0])*nVecLen);


		//ȷΪ0
		if ( nBound >= 1 ) {
			CirQueueBuf_Prev(1, pd_s, pd_y);
			lout_assert(VecZero(pd_s, m_pFunc->m_nParamNum) == false);
			lout_assert(VecZero(pd_y, m_pFunc->m_nParamNum) == false);
		}
		
		

		//һѭ
		for (int i=1; i<=nBound; i++)
		{
			CirQueueBuf_Prev(i, pd_s, pd_y);

			double dProd = VecProduct(pd_s, pd_y, nVecLen); 	
			pdAlpha[i-1] = VecProduct(pd_s, pdDir, nVecLen) / dProd;
			for (int n=0; n<nVecLen; n++)
				pdDir[n] -= pdAlpha[i-1] * pd_y[n];
		}


		//gamma,ʼH^0
		double dGamma = 1;
		if (k>=1) {
			CirQueueBuf_Prev(1, pd_s, pd_y);
			dGamma = VecProduct(pd_s, pd_y, nVecLen) / VecProduct(pd_y, pd_y, nVecLen);
		}

		// r = H^0 * q
		for (int n=0; n<nVecLen; n++)
			pdDir[n] *= dGamma;

		//ڶѭ
		for (int i=nBound; i>=1; i--)
		{
			CirQueueBuf_Prev(i, pd_s, pd_y);
			double dBeta = VecProduct(pd_y, pdDir, nVecLen) / VecProduct(pd_y, pd_s, nVecLen);
			for (int n=0; n<nVecLen; n++)
				pdDir[n] += pd_s[n] * (pdAlpha[i-1] - dBeta);
		}
		


		//Ҫȡ
		for (int n=0; n<nVecLen; n++)
			pdDir[n] = -pdDir[n];
	}
//#define SIMPLE_LINESEARCH
	/// һά㲽m_dOpt_FixedStep ֵ򲻽һά
	/** 
	 * \param [in] pdDir p_k
	 * \param [in] dValue x_kĺֵ
	 * \param [in] pdGradient x_kݶ
	 * \param [in] pdCurParam x_k
	 * \param [out] pdNextParam ͬʱx_(k+1)
	 * \return  ز
	 */
	virtual double LineSearch(double *pdDir, double dValue, double *pdGradient, double *pdCurParam, double *pdNextParam)
	{
		//m_dOpt_FixedStep ֵ򲻽һά
		if (m_opt_dFixedStep > 0) {
			for (int i=0; i<m_pFunc->m_nParamNum; i++)
				pdNextParam[i] = pdCurParam[i] + m_opt_dFixedStep * pdDir[i];
			return m_opt_dFixedStep;
		}
#ifdef SIMPLE_LINESEARCH
		/// 㷨μNumerical OptimizationP37Algorithm 3.1
		double dStep = 1.0;
		double dRoll = 0.9;
		double dConst = 1e-4 * VecProduct(pdGradient, pdDir, m_pFunc->m_nParamNum);

		while (1) 
		{
			for (int n=0; n<m_pFunc->m_nParamNum; n++)
				pdNextParam[n] = pdCurParam[n] + dStep * pdDir[n];

			double dNextValue = m_pFunc->GetValue(pdNextParam);
			double dLeft = dValue + dStep * dConst;

			if (dNextValue > dLeft)
				dStep *= dRoll;
			else
				break;
		}
		return dStep;
#else
		/// 㷨μNumerical OptimizationP57ܵinterpolation㷨
		double a0=0, a1=0, a2=0; //
		double phi0=0, phi1=0, phi2=0; //phi(ai)
		double c = 1e-4;
		double phi_t = VecProduct(pdGradient, pdDir, m_pFunc->m_nParamNum); // phi'(0)

		a2 = 1.0; //ʼΪ1
		for (int k=1; a2>0 ;k++)
		{
			// x = x + a * p
			for (int n=0; n<m_pFunc->m_nParamNum; n++)
				pdNextParam[n] = pdCurParam[n] + a2 * pdDir[n];

			phi2 = m_pFunc->GetValue(pdNextParam); // phi(a2)

			if (phi2 <= dValue + c * a2 * phi_t)
				break;
			
			//ǰεĽ
			a0 = a1;
			a1 = a2;
			phi0 = phi1;
			phi1 = phi2;
			if ( k== 1 ) {
				a2 = - phi_t*a1*a1 / 2 / ( phi1 - dValue - phi_t * a1 );
			} else {
				double v1 = phi1 - dValue - phi_t * a1;
				double v2 = phi0 - dValue - phi_t * a0;
				double a = a0*a0*v1 - a1*a1*v2;
				double b = -a0*a0*a0*v1 + a1*a1*a1*v2;
				double t = a0*a0*a1*a1*(a1-a0);
				a /= t;
				b /= t;
				a2 = (-b+sqrt(b*b-3*a*phi_t)) / (3*a);
			}

			//a2a1̫ӽ̫Զȡa2=a1/2
			if (fabs(a2-a1) < 1e-5 ||
				a1/a2 > 10 )
				a2 = a1/2;
			//a2Сȡһ̫Сֵ
			if (a2 < 1e-10) {
				lout_warning("a2 is too small < 1e-10, break")
				a2 = 1e-5;
				break;
			}

			if (m_opt_nLineSearchMaxIterNum>0 && k>m_opt_nLineSearchMaxIterNum) {
				a2 = 1e-6;
				break;
			}
		}
		return a2;
#endif
	}

	/// ڻ
	static double VecProduct(double *pdVec1, double *pdVec2, int nSize)
	{
		double d = 0;
		for (int i=0; i<nSize; i++) 
			d += pdVec1[i] * pdVec2[i];
		return d;
	}
	/// ģ
	static double VecNorm(double *pdVec, int nSize)
	{
		return sqrt(VecProduct(pdVec, pdVec, nSize));
	}
	/// ľ
	static double VecDist(double *pdVec1, double *pdVec2, int nSize)
	{
		double d = 0;
		for (int i=0; i<nSize; i++)
			d += pow(pdVec1[i]-pdVec2[i], 2);
		return sqrt(d);
	}
	/// жǷΪ0
	static bool VecZero(double *pdVec, int nSize)
	{
		for (int i=0; i<nSize; i++) {
			if (fabs(pdVec[i]) > 1e-10) 
				return false;
		}
		return true;
	}

	/// ָ
	void Repeat(double *&p1, double *&p2)
	{
		double *pTemp = p1;
		p1 = p2;
		p2 = pTemp;
	}
};