// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// option class.  Using to define the input of a system.
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


#ifndef _WB_OPTION_H_
#define _WB_OPTION_H_

#include "wbLog.h"

typedef enum {
	wbOPT_TRUE,			//ڣΪtrue
	wbOPT_FALSE,			//ڣΪfalse
	wbOPT_INT,			//
	wbOPT_STRING,			//ַ
	wbOPT_FLOAT			//ȸ
} ValueType;

typedef struct {
	ValueType type;			//
	const char* pLabel;		//ǩʹá-+pLabel"־Ҫlabelǰи-.
	void* pAddress;			//ַ
	const char* pDocMsg;	//
} wbOption;


#define wbOpt_Number(opts) (sizeof(opts)/sizeof(wbOption))
#define wbOpt_Parse(opts) wbOptParse::Parse(__argc, __argv, opts, (sizeof(opts)/sizeof(wbOption)))
#define wbOpt_ParseWithHelp(opts, strHelp) wbOptParse::Parse(__argc, __argv, opts, (sizeof(opts)/sizeof(wbOption)), strHelp)

class wbOptParse
{
public:
	//ӡ Usage
	static void PrintUsage(wbOption *pOptArray, int optNum)
	{
		for (int i=0; i<optNum; i++)
		{
			lout<<"-"<<pOptArray[i].pLabel<<"\t"<<pOptArray[i].pDocMsg<<endl;
		}
	}
	//ӡ value
	static void PrintValue(wbOption *pOptArray, int optNum)
	{
		for (int i=0; i<optNum; i++)
		{
			lout<<"-"<<pOptArray[i].pLabel<<" =\t";

			switch(pOptArray[i].type)
			{
			case wbOPT_TRUE:
				lout<<*( (bool*)(pOptArray[i].pAddress) );
				break;
			case wbOPT_FALSE:
				lout<<*( (bool*)(pOptArray[i].pAddress) );
				break;
			case wbOPT_INT:
				lout<<*( (int*)(pOptArray[i].pAddress) );
				break;
			case wbOPT_FLOAT:
				lout<<*( (float*)(pOptArray[i].pAddress) );
				break;
			case wbOPT_STRING:
				lout<<*( (char**)(pOptArray[i].pAddress) );
				break;
			}

			lout<<endl;
		}
	}
	//ʼͲͬ費ͬĳֵ
	static void Init(wbOption *pOptArray, int optNum)
	{
		for (int i=0; i<optNum; i++)
		{
			wbOption *pOpt = &pOptArray[i];
			switch(pOptArray[i].type)
			{
			case wbOPT_TRUE:
				*( (bool*)(pOptArray[i].pAddress) ) = false;
				break;
			case wbOPT_FALSE:
				*( (bool*)(pOptArray[i].pAddress) ) = true;
				break;
			case wbOPT_INT:
				*( (int*)(pOptArray[i].pAddress) ) = 0;
				break;
			case wbOPT_FLOAT:
				*( (float*)(pOptArray[i].pAddress) ) = 0;
				break;
			case wbOPT_STRING:
				*( (char**)(pOptArray[i].pAddress) ) = NULL;
				break;
			}
		}
	}
	///< з
	/** 
	 * \param[in] argc Ŀ
	 * \param[in] argv ַ
	 * \param[in] pOptArray wbOption
	 * \param[in] optNum wbOptionĿ
	 * \param[in] pOtherHelp ȡʧܣӡpOtherHelpݣ
	 * \param[in] bOutValue Ƿӡÿֵ
	 * \param[in] bMustCommand ΪtrueûΪȡʧ
	*/
	static int Parse(int argc, char **argv, wbOption *pOptArray, int optNum, 
		const char *pOtherHelp=NULL, bool bOutValue = true, bool bMustCommand = true)
	{
		if (argc == 1 && bMustCommand) {
			PrintUsage(pOptArray, optNum);
			if (pOtherHelp)
				lout<<pOtherHelp<<endl;
			lout<<"..."<<endl;
			getch();
 			exit(0);
		}

		register int CommandCur = 0;
		register char *CommandPtr = NULL;
		register int OptCur = 0;
		register wbOption *OptPtr = NULL;

		int nActiveOptNum = 0;
		//ʼʼⲿֶ
		//Init(pOptArray, optNum);
		for (CommandCur=1; CommandCur<argc; )
		{
			CommandPtr = argv[CommandCur];
			if (*CommandPtr == '-') //label
			{
				OptCur = 0;
				for(; OptCur<optNum; OptCur++)
				{
					if ( 0 == strcmp(CommandPtr+1, pOptArray[OptCur].pLabel) )
						break;
				}

				if (OptCur >= optNum) //δ֪Label
				{
					if (strcmp(CommandPtr+1, "?") == 0) { // ʹ˵
						PrintUsage(pOptArray, optNum);
						exit(0);
					} else if (strcmp(CommandPtr+1, "log") == 0) {//logļ޸
						if (CommandCur+1 < argc) {
							lout.ReFile(argv[CommandCur+1], true);
							CommandCur+=2;
						} else {
							lout_error("-log Чļ!");
						}
					} else {
						lout_error("Unknown label: "<<CommandPtr);
					}
				}
				else
				{
					OptPtr = &pOptArray[OptCur];

					if (OptPtr->type != wbOPT_TRUE && OptPtr->type != wbOPT_FALSE)
					{
						CommandCur++;
						if (CommandCur >= argc)
						{
							cout<<"[ERROR]\t"<<CommandPtr<<"ûжӦvalue"<<endl;
							exit(0);
						}

						CommandPtr = argv[CommandCur];
					}

					switch(OptPtr->type)
					{
					case wbOPT_TRUE:
						*( (bool*)(OptPtr->pAddress) ) = true;
						break;
					case wbOPT_FALSE:
						*( (bool*)(OptPtr->pAddress)  ) = false;
						break;
					case wbOPT_INT:
						*( (int*)(OptPtr->pAddress)  ) = atoi(CommandPtr);
						break;
					case wbOPT_FLOAT:
						*( (float*)(OptPtr->pAddress) ) = atof(CommandPtr);
						break;
					case wbOPT_STRING:
						*( (char**)(OptPtr->pAddress) ) = CommandPtr;
						break;
					}

					CommandCur++;

					nActiveOptNum++;
				}

			}
			else
			{
				cout<<"Unknown label: "<<CommandPtr<<endl;
				PrintUsage(pOptArray, optNum);
				exit(0);
			}
		}

		if (bOutValue)
			PrintValue(pOptArray, optNum);

		return nActiveOptNum;
	}

	///< ļзļʽÿΪ[label] = [value]еȺ˵Ŀոп
	/** 
	 * \param[in] argc Ŀ
	 * \param[in] argv ַ
	 * \param[in] pOptArray wbOption
	 * \param[in] optNum wbOptionĿ
	 * \param[in] pOtherHelp ȡʧܣӡpOtherHelpݣ
	 * \param[in] bOutValue Ƿӡÿֵ
	 * \param[in] bMustCommand ΪtrueûΪȡʧ
	*/
// 	static int Parse(const char *pathOption, wbOption *pOptArray, int optNum, 
// 		const char *pOtherHelp=NULL, bool bOutValue = true, bool bMustCommand = true) 
// 	{
// 		FILE *fp = NULL;
// 		if (pathOption) {
// 			fp = fopen(pathOption, "rt");
// 		}
// 
// 		int nOptLine = 0;
// 		char strLine[1024*32];
// 		while ( fgets(strLine, 1024*32, fp) ) {
// 			nOptLine++;
// 
// 			int nLen = strlen(strLine);
// 			if (strLine(nLen-1) != '\n') {
// 				lout_error("[OptFile] nLine="<<nOptLine<<": Line too long!");
// 			}
// 
// 
// 		}
// 	}
};


#endif