// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// Memory pool class
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


#ifndef _WB_POOL_H_
#define _WB_POOL_H_

#include "wbVector.h"
#include "wbLog.h"
/*
ڴ
*/

template <class Type>
class wbUnitPool
{
public:
	//static wbStack<Type*> pointerStack;

	static wbUnitPool<Type> *_head;
	static wbUnitPool<Type> *_tail;

	wbUnitPool<Type> *_next;

#ifdef _DEBUG
	//static wbArray<Type*> memArray;
	static unsigned int nNewNum;
	static unsigned int nDelNum;
	static unsigned int nMemNum; //ܹʹnew˶ٸַ
#endif

	inline static Type* Get()
	{

#ifdef _DEBUG
		nNewNum++;
#endif

		if (!_head) {
#ifdef _DEBUG
			nMemNum++;
#endif
			return NULL;
		}
		Type *pBuf = (Type*)_head;
		_head = _head->_next;
		if (!_head)
			_tail = NULL;

		return pBuf;
	}

	inline static Type *New()
	{
		Type *p = Get();

		if (p) {
			new (p) Type;
		} else {
			p = new Type;
			//memArray.Add(p);
		}

		return p;
	}

	inline static void Delete(Type *&p)
	{
#ifdef _DEBUG
		nDelNum++;
#endif
		if (!_head)
		{
			_head = (wbUnitPool<Type> *)p;
			_tail = _head;
		} 
		else
		{
			_tail->_next = (wbUnitPool<Type> *)p;
			_tail = _tail->_next;
		}
		_tail->_next = NULL;
	}

	inline static void ReleaseAll()
	{
#ifdef _DEBUG
		nNewNum=0;
		nDelNum=0;
		nMemNum = 0;
		//memArray.Clean();
#endif
// 		Type *p;
// 		while (pointerStack.Pop(&p)) {
// 			delete p;
// 		}
		while (_head)
		{
			wbUnitPool<Type> *p = _head;
			_head = _head->_next;
			delete p;
		}
		_head = NULL;
		_tail = NULL;
	}
};

//template <class Type>
//wbStack<Type*> wbUnitPool<Type>::pointerStack;
template <class Type>
wbUnitPool<Type> *wbUnitPool<Type>::_head = NULL;
template <class Type>
wbUnitPool<Type> *wbUnitPool<Type>::_tail = NULL;
#ifdef _DEBUG
//template<class Type>
//wbArray<Type*> wbUnitPool<Type>::memArray;
template<class Type>
unsigned int wbUnitPool<Type>::nNewNum = 0;
template<class Type>
unsigned int wbUnitPool<Type>::nDelNum = 0;
template<class Type>
unsigned int wbUnitPool<Type>::nMemNum = 0;
#endif

/*
ʹҪʱض̳д
ʹ֮ǰҪʼ
ʹúҪͷڴ
*/
template <class Type>
class wbListPool
{
public:
	wbListPool<Type> *_prev;
	wbListPool<Type> *_next;
	
	static wbListPool<Type> *_head;
	static wbListPool<Type> *_tail;

	static int _active_num; //ǰʹõڴĿ
	static int _memseg_num; //ܹٵڴ

	/*
	˴ָܶгʼΪͨNewڴҪָϢ,ⲿ³ʼὫָϢ
	*/
	wbListPool(){};

	static void Init()
	{
		if (!_head || !_tail) //ʼһ߾Ϳ
		{
			//һӵ㣬ʼͽڵ
			Type *pH = new Type;
			_head = (wbListPool<Type> *)pH;
			_tail = _head;
			_head->_prev = NULL;
			_head->_next = NULL;

			_active_num = 0;
			_memseg_num = 0;
		}
	}
	static void Release()
	{
		//ͷŵڴ
		for (wbListPool<Type> *p = _head; p; ) {
			Type *pType = (Type*)p;
			p = p->_next;
			delete pType;
		}

		_head = NULL;
		_tail = NULL;
		_active_num = 0;
		_memseg_num = 0;
	}
	//NewлĬϹ캯˿ܻӰЧʣʱҪֶGetAddֹظù캯
	static Type *Get() //ӳлڴ棬Ϊգ򷵻NULL
	{
		if (!_head || !_tail)
			lout_error("wbListPool ʹ֮ǰδгʼ");

		_active_num++;

		Type *pBuf;
		if (_tail->_next) {
			pBuf = (Type*)(_tail->_next);
			_tail = _tail->_next;

			return pBuf;

		}

		return NULL;
	}
	static void Add(Type *pBuf) //һ¿ٵڴӵ
	{
		_memseg_num++;

		wbListPool<Type> *p = (wbListPool<Type> *)pBuf;


		//Tailĺ
		p->_next = _tail->_next;
		p->_prev = _tail;
		_tail->_next = p;
		if (p->_next)
			p->_next->_prev = p;
		_tail = p;
	}
	static Type *New()
	{
		Type *pBuf = Get();
		if (pBuf) {
			new (pBuf) Type;
		} else {
			pBuf = new Type;
			Add(pBuf);
		}

		return pBuf;
	}

	static void Delete(Type *pBuf)
	{
		_active_num--;

		//pBuf->~Type();
		wbListPool<Type> *p = (wbListPool<Type> *)pBuf;

		if (p == _tail) {
			_tail = _tail->_prev;
			return;
		}


		p->_next->_prev = p->_prev;
		p->_prev->_next = p->_next;

		p->_next = _tail->_next;
		p->_prev = _tail;
		_tail->_next = p;
		if (p->_next)
			p->_next->_prev = p;
	}

	static void DeleteAll()
	{
		_active_num = 0;

		//ֻûֶͷڴʱſֱӹλָ

		_tail = _head;
	}

	static void MoveToHead(Type *pBuf)
	{
		if (!pBuf)
			return;

		wbListPool<Type> *p = (wbListPool<Type> *)pBuf;
		//жǷǱɾڴ
		for (wbListPool<Type> *pTemp = p; ; pTemp=pTemp->_prev)
		{
			if (pTemp->_prev == _tail) //˵ǱɾĽڵ㣬
				return;
			if (pTemp->_prev == _head) //Чڵ
				break;
		}

		if (_tail == _head->_next)
			return; //ֻһڵ㣬ƶ

		if (_tail == p) //pǽβ
			_tail = p->_prev;

		p->_prev->_next = p->_next;
		if (p->_next)
			p->_next->_prev = p->_prev;

		p->_next = _head->_next;
		_head->_next->_prev = p;
		p->_prev = _head;
		_head->_next = p;
	}
};

template<class Type>
wbListPool<Type> *wbListPool<Type>::_head = NULL;
template<class Type>
wbListPool<Type> *wbListPool<Type>::_tail = NULL;
template<class Type>
int wbListPool<Type>::_active_num = 0;
template<class Type>
int wbListPool<Type>::_memseg_num = 0;



#endif