// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2014-2015 Tsinghua University
// Author: wb.th08@gmail.com (Bin Wang), ozj@tsinghua.edu.cn (Zhijian Ou) 
//
// \file
// System file, including many simple definition.
// All h, cpp, cc, and script files (e.g. bat, sh, pl, py) should include the above 
// license declaration. Different coding language may use different comment styles.


#ifndef _WB_SYSTEM_H_
#define _WB_SYSTEM_H_


#include <ctime>
#include <cstring>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdio>
#include <cstring>
#include <cstdarg>	//va_listԶɱprint
#include <conio.h>  //getch() //ԲԵĽַ
using namespace std;
#include "wbLog.h"
#include "wbVector.h"
#include "wbPool.h"
#include "wbOption.h"
#include "wbFile.h"
#include "wbLargeFile.h"
#include "wbString.h"
#include "wbLHash.h"
#include "wbTrie.h"
#include "wbHeap.h"
#include "wbChTxt.h"


#ifdef UNICODE
#error "ʹöֽڵĻ롣unicodeĵʵͶд"
#endif

//===================  ================================
#define GS_SENTENCE_LEN	32*1024 // 5000

//==================== Main  ===========================
#define _wbMain int main(int _argc, char** _argv)
//====================Debug ================================
#define F_RETURN(b) {if(!(b)) {cout<<"<F_RETURN>Error = "<<__FILE__<<" (line "<<__LINE__<<")"<<endl; return 0;} }
#define ERROR(b) { cout<<b<<endl; return 0; }
//=====================ļ===================================
// #define SAFE_FOPEN(pfile, path, mode) {\
// 	if ( (pfile = fopen(path, mode)) == NULL ) {cout<<"File Open Failed: "<<path<<endl; exit(0);}}
// 
// #define SAFE_FMOPEN(file, path, mode) {\
// 	file.open(path, mode); if (!file.good()) {cout<<"File Open Failed: "<<path<<endl; exit(0);}}
// 
// #define SAFE_FSOPEN(file, path) {\
// 	file.open(path); if (!file.good()) {cout<<"File Open Failed: "<<path<<endl; exit(0);}}
// //=====================ļر===================================
// #define SAFE_FCLOSE(fp) { if(fp) {fclose(fp); fp=NULL;} }

//=====================ڴ濪=============================
#define POINT_TEST(p) { if ((p)==NULL) {cout<<#p<<"[New Memory Error]"<<endl; exit(0);} }
#define SAFE_NEW(p, Type) {p = new Type; POINT_TEST(p)}
#define SAFE_DNEW(PType, p, Type) { PType p = new Type; POINT_TEST(p) }
#define SAFE_NEW_ARRAY(p, Type, n) { p = new Type[n]; POINT_TEST(p) }
#define SAFE_NEW_DARRAY(p, Type, n, m) { p=new Type*[n]; for(int i=0; i<n; i++){p[i]=new Type[m];}; POINT_TEST(p) }
//=====================ڴͷ=============================
#define SAFE_DELETE(p) {if (p) delete (p); (p)=NULL;}
#define SAFE_DELETE_ARRAY(p) {if(p) delete [](p); (p)=NULL;}
#define SAFE_DELETE_DARRAY(p, n) { if(p){ for(int i=0; i<n; i++){delete [](p)[i];} delete [](p); (p)=NULL; } }

//====================Դӡ==============================
#define Value_Print(value) {cout<<#value" = "<<value<<endl;}
#define wbFile_Value_Print(file, value, form) { file.Print(#value" = "#form"\n", value); }

//=================== static  ==========================



//===================== typedef =============================
//typedef char PATH[256];
typedef char PATH_PACK[256*10];	//·

//================= class PATH =============================
//printfֺҪӡPATHݣֱӴҪǿƽPATHתchar*ͣӡ
//wbString࣬Ҳᷢͬ
class PATH
{
public:
	char buffer[256];
	PATH() { buffer[0] = '\0'; }
	PATH(const char *p_str) { strcpy(buffer, p_str); }

	bool good() { return buffer[0]!='\0'; } 
	operator char* () { return buffer; }
	char *FileName() { char *p = strrchr(buffer, '\\'); return (p)? p+1:buffer; }
};

//=========================================================
//===============  ࣺ wbPath =========================
//========================================================
#define MAX_PATH_LEN 256
#define MAX_PATH_NUM 10
#define MAX_PATHPACK_LEN (MAX_PATH_LEN*MAX_PATH_NUM+100) //ǵPathPackл
class wbPath
{
public:
	char m_pathPack[MAX_PATHPACK_LEN];
	char* m_pPath[MAX_PATH_NUM];	//ָÿpath<ŵpath>
	int m_nPathNum;
	wbQueue<char*> *m_pPathQueue;	//ҵpath
	wbPath();
	wbPath(char *p_path);
	~wbPath();

	char* GetBuffer() { return m_pathPack; }
	
	void pathUnPack(); //Ϊm_pPathֵ
	void pathSearch(); //Ϊm_pPathQueueֵ
	
	bool GetPath( char *p_path ); //Իpath

	void LoadListFile( char *p_pBasePath = "", char *p_pSuffix="", char *p_pListFile = NULL); //Listļfilepath
};

//=========================================================
//==========  ࣺConfig   =============================
//=========================================================
class wbConfig_Value 
{
public:
	void *pValue;
	const char *pFormat;
	const char *pDepiction;

	bool bExist;

	wbConfig_Value()
	{
		pValue = NULL;
		pFormat = NULL;
		pDepiction = NULL;
		bExist = false;
	}
};

#define CONFIG_VALUE_NUM 50
class wbConfig
{
public:
	int m_nConfigNum;
	char** m_ppConfigVec;

	bool m_bCopy;

	wbConfig_Value *m_pValue;
	int m_nValueNum;

	const char* m_pProgramDepiction;
public:
	wbConfig(int p_ncNum, char** p_ppCVec, bool p_bCopy=false); //bCopyΪtrueڲڴ洢
	~wbConfig(void);
	bool ScanConfig(const char* p_format, void* p_pValue=NULL);
	void PrintConfig(ostream &p_out = cout);
	
	//bool ScanFileList(const char* p_label, wbVector<char*>* pVector);
	//static bool ScanFileList(char* p_pFileStr, wbVector<char*>* pVector);

	void ValueLabel(const char* p_pFormat, void* p_pValue = NULL, const char* p_pDepiction = NULL);
	bool ValueSet();
	void ValuePrint();
	void ValueUsage();
	bool ValueExist(const char* p_label);
};

//=========================================================
//==========  ࣺʱ    =============================
//=========================================================
class wbClock
{
public:
	clock_t m_nBeginTime;
	clock_t m_nEndTime;
	bool m_bWork;
public:
	wbClock(void);
	~wbClock(void);
	void Clean();
	clock_t Begin();
	clock_t End();
	clock_t Get(); //õǰʱ䣬ֹͣʱ

	static double ToSecond(clock_t t) { return 1.0*t/1000; }
};


//=====================ٷֺ==================================
#define MAX_CMD_TITLE_LEN 1000
void outPrecent(long long n, bool bNew = false, long long nTotal = 100, const char* title = "Process");
void outPrecent(ifstream &ifile, bool bNew = false, const char* title = "Process");
void outPrecent(FILE *fp, bool bNew = false, const char* title = "Process");
/// nState: 0, ֵ1ʼֵ2趨title
void titlePrecent(long long n = -1, short nState = 0, long long nTotal = 100, const char* title = "Process");
void titlePrecent(ifstream &ifile, bool bNew = false, const char* title = "Process");
void titlePrecent(FILE *fp, bool bNew = false, const char* title = "Process");


//======================йغֱĶ========================
#define GS_CHAR_BASE_ZH 32768  //ԴΪֱ뵽ֱ֤3ʼ0/1/2Ϊ<eps><s></s>
#define GS_CHAR_CEILING_ZH 32768
unsigned Char2Num(const char *info);
void Num2Char(unsigned n, char *info);
bool bChSymbol(const char *str);
bool bChCharSeq(const char *str); //ַǷǺִӢģַ
//const char c_aChNumChar[10][3] = {"", "һ", "", "", "", "", "", "", "", ""}

//򵥲
void Pause(); /// ͣ
void Title(const char* p);

#endif